# sage_setup: distribution = sagemath-categories
# sage.doctest: needs sage.combinat sage.groups
r"""
Weyl Groups
"""
# ****************************************************************************
#  Copyright (C) 2009    Nicolas M. Thiery <nthiery at users.sf.net>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  https://www.gnu.org/licenses/
# *****************************************************************************

from sage.misc.cachefunc import cached_method, cached_in_parent_method
from sage.misc.lazy_import import LazyImport
from sage.categories.category_singleton import Category_singleton
from sage.categories.coxeter_groups import CoxeterGroups


class WeylGroups(Category_singleton):
    r"""
    The category of Weyl groups.

    See the :wikipedia:`Wikipedia page of Weyl Groups <Weyl_group>`.

    EXAMPLES::

        sage: WeylGroups()
        Category of Weyl groups
        sage: WeylGroups().super_categories()
        [Category of Coxeter groups]

    Here are some examples::

        sage: WeylGroups().example()            # todo: not implemented
        sage: FiniteWeylGroups().example()
        The symmetric group on {0, ..., 3}
        sage: AffineWeylGroups().example()      # todo: not implemented
        sage: WeylGroup(["B", 3])
        Weyl Group of type ['B', 3] (as a matrix group acting on the ambient space)

    This one will eventually be also in this category::

        sage: SymmetricGroup(4)
        Symmetric group of order 4! as a permutation group

    TESTS::

        sage: C = WeylGroups()
        sage: TestSuite(C).run()
    """

    def super_categories(self):
        r"""
        EXAMPLES::

            sage: WeylGroups().super_categories()
            [Category of Coxeter groups]
        """
        return [CoxeterGroups()]

    def additional_structure(self):
        r"""
        Return  ``None``.

        Indeed, the category of Weyl groups defines no additional
        structure: Weyl groups are a special class of Coxeter groups.

        .. SEEALSO:: :meth:`Category.additional_structure`

        .. TODO:: Should this category be a :class:`CategoryWithAxiom`?

        EXAMPLES::

            sage: WeylGroups().additional_structure()
        """
        return None

    Finite = LazyImport('sage.categories.finite_weyl_groups', 'FiniteWeylGroups')

    class ParentMethods:
        def coxeter_matrix(self):
            """
            Return the Coxeter matrix associated to ``self``.

            EXAMPLES::

                sage: G = WeylGroup(['A',3])
                sage: G.coxeter_matrix()
                [1 3 2]
                [3 1 3]
                [2 3 1]
            """
            return self.cartan_type().coxeter_matrix()

        def pieri_factors(self, *args, **keywords):
            r"""
            Return the set of Pieri factors in this Weyl group.

            For any type, the set of Pieri factors forms a lower ideal
            in Bruhat order, generated by all the conjugates of some
            special element of the Weyl group. In type `A_n`, this
            special element is `s_n\cdots s_1`, and the conjugates are
            obtained by rotating around this reduced word.

            These are used to compute Stanley symmetric functions.

            .. SEEALSO::

                * :meth:`WeylGroups.ElementMethods.stanley_symmetric_function`
                * :mod:`sage.combinat.root_system.pieri_factors`

            EXAMPLES::

                sage: W = WeylGroup(['A',5,1])
                sage: PF = W.pieri_factors()
                sage: PF.cardinality()
                63

                sage: W = WeylGroup(['B',3])
                sage: PF = W.pieri_factors()
                sage: sorted([w.reduced_word() for w in PF])
                [[],
                 [1],
                 [1, 2],
                 [1, 2, 1],
                 [1, 2, 3],
                 [1, 2, 3, 1],
                 [1, 2, 3, 2],
                 [1, 2, 3, 2, 1],
                 [2],
                 [2, 1],
                 [2, 3],
                 [2, 3, 1],
                 [2, 3, 2],
                 [2, 3, 2, 1],
                 [3],
                 [3, 1],
                 [3, 1, 2],
                 [3, 1, 2, 1],
                 [3, 2],
                 [3, 2, 1]]
                sage: W = WeylGroup(['C',4,1])
                sage: PF = W.pieri_factors()
                sage: W.from_reduced_word([3,2,0]) in PF
                True
            """
            # Do not remove this line which makes sure the pieri factor
            # code is properly inserted inside the Cartan Types
            import sage.combinat.root_system.pieri_factors
            assert sage.combinat.root_system.pieri_factors
            ct = self.cartan_type()
            if hasattr(ct, "PieriFactors"):
                return ct.PieriFactors(self, *args, **keywords)
            raise NotImplementedError("Pieri factors for type {}".format(ct))

        def bruhat_cone(self, x, y, side='upper', backend='cdd'):
            r"""
            Return the (upper or lower) Bruhat cone associated to the interval ``[x,y]``.

            To a cover relation `v \prec w` in strong Bruhat order you can assign a positive
            root `\beta` given by the unique reflection `s_\beta` such that `s_\beta v = w`.

            The upper Bruhat cone of the interval `[x,y]` is the non-empty, polyhedral cone generated
            by the roots corresponding to `x \prec a` for all atoms `a` in the interval.
            The lower Bruhat cone of the interval `[x,y]` is the non-empty, polyhedral cone generated
            by the roots corresponding to `c \prec y` for all coatoms `c` in the interval.

            INPUT:

            - ``x`` -- an element in the group `W`

            - ``y`` -- an element in the group `W`

            - ``side`` -- (default: ``'upper'``) must be one of the following:

              * ``'upper'`` -- return the upper Bruhat cone of the interval [``x``, ``y``]
              * ``'lower'`` -- return the lower Bruhat cone of the interval [``x``, ``y``]

            - ``backend`` -- string (default: ``'cdd'``); the backend to use to create the polyhedron

            EXAMPLES::

                sage: W = WeylGroup(['A',2])
                sage: x = W.from_reduced_word([1])
                sage: y = W.w0
                sage: W.bruhat_cone(x, y)
                A 2-dimensional polyhedron in QQ^3
                 defined as the convex hull of 1 vertex and 2 rays

                sage: W = WeylGroup(['E',6])
                sage: x = W.one()
                sage: y = W.w0
                sage: W.bruhat_cone(x, y, side='lower')
                A 6-dimensional polyhedron in QQ^8
                 defined as the convex hull of 1 vertex and 6 rays

            TESTS::

                sage: W = WeylGroup(['A',2])
                sage: x = W.one()
                sage: y = W.w0
                sage: W.bruhat_cone(x, y, side='nonsense')
                Traceback (most recent call last):
                ...
                ValueError: side must be either 'upper' or 'lower'

            REFERENCES:

            - [Dy1994]_
            - [JS2021]_
            """
            from sage.modules.free_module_element import vector
            if side == 'upper':
                roots = [vector((x * r * x.inverse()).reflection_to_root().to_ambient())
                         for z, r in x.bruhat_upper_covers_reflections()
                         if z.bruhat_le(y)]
            elif side == 'lower':
                roots = [vector((y * r * y.inverse()).reflection_to_root().to_ambient())
                         for z, r in y.bruhat_lower_covers_reflections()
                         if x.bruhat_le(z)]
            else:
                raise ValueError("side must be either 'upper' or 'lower'")

            from sage.geometry.polyhedron.constructor import Polyhedron
            return Polyhedron(vertices=[vector([0] * self.degree())],
                              rays=roots,
                              ambient_dim=self.degree(),
                              backend=backend)

        @cached_method
        def quantum_bruhat_graph(self, index_set=()):
            r"""
            Return the quantum Bruhat graph of the quotient of the Weyl
            group by a parabolic subgroup `W_J`.

            INPUT:

            - ``index_set`` -- (default: ()) a tuple `J` of nodes of
              the Dynkin diagram

            By default, the value for ``index_set`` indicates that the
            subgroup is trivial and the quotient is the full Weyl group.

            EXAMPLES::

                sage: W = WeylGroup(['A',3], prefix='s')
                sage: g = W.quantum_bruhat_graph((1,3))
                sage: g
                Parabolic Quantum Bruhat Graph of Weyl Group of type ['A', 3]
                 (as a matrix group acting on the ambient space)
                 for nodes (1, 3): Digraph on 6 vertices
                sage: g.vertices(sort=True)
                [s2*s3*s1*s2, s3*s1*s2, s1*s2, s3*s2, s2, 1]
                sage: g.edges(sort=True)
                [(s2*s3*s1*s2, s2, alpha[2]),
                 (s3*s1*s2, s2*s3*s1*s2, alpha[1] + alpha[2] + alpha[3]),
                 (s3*s1*s2, 1, alpha[2]),
                 (s1*s2, s3*s1*s2, alpha[2] + alpha[3]),
                 (s3*s2, s3*s1*s2, alpha[1] + alpha[2]),
                 (s2, s1*s2, alpha[1] + alpha[2]),
                 (s2, s3*s2, alpha[2] + alpha[3]),
                 (1, s2, alpha[2])]
                sage: W = WeylGroup(['A',3,1], prefix='s')
                sage: g = W.quantum_bruhat_graph()
                Traceback (most recent call last):
                ...
                ValueError: the Cartan type ['A', 3, 1] is not finite
            """
            if not self.cartan_type().is_finite():
                raise ValueError("the Cartan type {} is not finite".format(self.cartan_type()))

            # This is a modified form of quantum_bruhat_successors.
            # It does not do any error checking and also is more efficient
            #   with how it handles memory and checks by using data stored
            #   at this function level rather than recomputing everything.
            lattice = self.cartan_type().root_system().root_lattice()
            NPR = lattice.nonparabolic_positive_roots(index_set)
            NPR_sum = sum(NPR)
            NPR_data = {}
            double_rho = lattice.sum(lattice.positive_roots())  # = 2 * \rho
            for alpha in NPR:
                ref = alpha.associated_reflection()
                alphacheck = alpha.associated_coroot()
                NPR_data[alpha] = [self.from_reduced_word(ref),  # the element
                                   len(ref) == double_rho.scalar(alphacheck) - 1,  # is_quantum
                                   NPR_sum.scalar(alphacheck)]  # the scalar
            # We also create a temporary cache of lengths as they are
            #   relatively expensive to compute and needed frequently
            visited = {}
            todo = {self.one()}
            len_cache = {}

            def length(x):
                if x in len_cache:
                    return len_cache[x]
                len_cache[x] = x.length()
                return len_cache[x]
            while todo:
                x = todo.pop()
                w_length_plus_one = length(x) + 1
                adj = {}
                for alpha in NPR:
                    elt, is_quantum, scalar = NPR_data[alpha]
                    wr = x * elt
                    wrc = wr.coset_representative(index_set)
                    # coset_representative returns wr if nothing gets changed
                    if wrc is wr and length(wrc) == w_length_plus_one:
                        if wrc not in visited:
                            todo.add(wrc)
                        adj[wr] = alpha
                    elif is_quantum and length(wrc) == w_length_plus_one - scalar:
                        if wrc not in visited:
                            todo.add(wrc)
                        adj[wrc] = alpha
                visited[x] = adj

            from sage.graphs.digraph import DiGraph
            return DiGraph(visited,
                           name="Parabolic Quantum Bruhat Graph of %s for nodes %s" % (self, index_set),
                           format='dict_of_dicts',
                           data_structure='static_sparse')

    class ElementMethods:

        def is_pieri_factor(self):
            r"""
            Return whether ``self`` is a Pieri factor, as used for
            computing Stanley symmetric functions.

            .. SEEALSO::

                * :meth:`stanley_symmetric_function`
                * :meth:`WeylGroups.ParentMethods.pieri_factors`

            EXAMPLES::

                sage: W = WeylGroup(['A',5,1])
                sage: W.from_reduced_word([3,2,5]).is_pieri_factor()
                True
                sage: W.from_reduced_word([3,2,4,5]).is_pieri_factor()
                False

                sage: W = WeylGroup(['C',4,1])
                sage: W.from_reduced_word([0,2,1]).is_pieri_factor()
                True
                sage: W.from_reduced_word([0,2,1,0]).is_pieri_factor()
                False

                sage: W = WeylGroup(['B',3])
                sage: W.from_reduced_word([3,2,3]).is_pieri_factor()
                False
                sage: W.from_reduced_word([2,1,2]).is_pieri_factor()
                True
            """

            return self in self.parent().pieri_factors()

        def left_pieri_factorizations(self, max_length=None):
            r"""
            Return all factorizations of ``self`` as `uv`, where `u`
            is a Pieri factor and `v` is an element of the Weyl group.

            .. SEEALSO::

                * :meth:`WeylGroups.ParentMethods.pieri_factors`
                * :mod:`sage.combinat.root_system.pieri_factors`

            EXAMPLES:

            If we take `w = w_0` the maximal element of a strict parabolic
            subgroup of type `A_{n_1} \times \cdots \times A_{n_k}`, then the Pieri
            factorizations are in correspondence with all Pieri factors, and
            there are `\prod 2^{n_i}` of them::

                sage: W = WeylGroup(['A', 4, 1])
                sage: W.from_reduced_word([]).left_pieri_factorizations().cardinality()
                1
                sage: W.from_reduced_word([1]).left_pieri_factorizations().cardinality()
                2
                sage: W.from_reduced_word([1,2,1]).left_pieri_factorizations().cardinality()
                4
                sage: W.from_reduced_word([1,2,3,1,2,1]).left_pieri_factorizations().cardinality()
                8

                sage: W.from_reduced_word([1,3]).left_pieri_factorizations().cardinality()
                4
                sage: W.from_reduced_word([1,3,4,3]).left_pieri_factorizations().cardinality()
                8

                sage: W.from_reduced_word([2,1]).left_pieri_factorizations().cardinality()
                3
                sage: W.from_reduced_word([1,2]).left_pieri_factorizations().cardinality()
                2
                sage: [W.from_reduced_word([1,2]).left_pieri_factorizations(max_length=i).cardinality()
                ....:  for i in [-1, 0, 1, 2]]
                [0, 1, 2, 2]

                sage: W = WeylGroup(['C',4,1])
                sage: w = W.from_reduced_word([0,3,2,1,0])
                sage: w.left_pieri_factorizations().cardinality()
                7
                sage: [(u.reduced_word(),v.reduced_word())
                ....:  for (u,v) in w.left_pieri_factorizations()]
                [([], [3, 2, 0, 1, 0]),
                ([0], [3, 2, 1, 0]),
                ([3], [2, 0, 1, 0]),
                ([3, 0], [2, 1, 0]),
                ([3, 2], [0, 1, 0]),
                ([3, 2, 0], [1, 0]),
                ([3, 2, 0, 1], [0])]

                sage: W = WeylGroup(['B',4,1])
                sage: W.from_reduced_word([0,2,1,0]).left_pieri_factorizations().cardinality()
                6
            """
            if max_length is None:
                from sage.rings.infinity import infinity
                max_length = infinity
            pieri_factors = self.parent().pieri_factors()

            def predicate(u):
                return u in pieri_factors and u.length() <= max_length

            return self.binary_factorizations(predicate)

        @cached_in_parent_method
        def stanley_symmetric_function_as_polynomial(self, max_length=None):
            r"""
            Return a multivariate generating function for the number
            of factorizations of a Weyl group element into Pieri
            factors of decreasing length, weighted by a statistic on
            Pieri factors.

            .. SEEALSO::

                * :meth:`stanley_symmetric_function`
                * :meth:`WeylGroups.ParentMethods.pieri_factors`
                * :mod:`sage.combinat.root_system.pieri_factors`

            INPUT:

                - ``self`` -- an element `w` of a Weyl group `W`
                - ``max_length`` -- nonnegative integer or infinity (default: infinity)

            Returns the generating series for the Pieri factorizations
            `w = u_1 \cdots u_k`, where `u_i` is a Pieri factor for
            all `i`, `l(w) = \sum_{i=1}^k l(u_i)` and
            ``max_length`` `\geq l(u_1) \geq \cdots \geq l(u_k)`.

            A factorization `u_1 \cdots u_k` contributes a monomial of
            the form `\prod_i x_{l(u_i)}`, with coefficient given by
            `\prod_i 2^{c(u_i)}`, where `c` is a type-dependent
            statistic on Pieri factors, as returned by the method
            ``u[i].stanley_symm_poly_weight()``.

            EXAMPLES::

                sage: W = WeylGroup(['A', 3, 1])
                sage: W.from_reduced_word([]).stanley_symmetric_function_as_polynomial()
                1
                sage: W.from_reduced_word([1]).stanley_symmetric_function_as_polynomial()
                x1
                sage: W.from_reduced_word([1,2]).stanley_symmetric_function_as_polynomial()
                x1^2
                sage: W.from_reduced_word([2,1]).stanley_symmetric_function_as_polynomial()
                x1^2 + x2
                sage: W.from_reduced_word([1,2,1]).stanley_symmetric_function_as_polynomial()
                2*x1^3 + x1*x2
                sage: W.from_reduced_word([1,2,1,0]).stanley_symmetric_function_as_polynomial()
                3*x1^4 + 2*x1^2*x2 + x2^2 + x1*x3
                sage: x = W.from_reduced_word([1,2,3,1,2,1,0])
                sage: x.stanley_symmetric_function_as_polynomial()  # long time
                22*x1^7 + 11*x1^5*x2 + 5*x1^3*x2^2 + 3*x1^4*x3 + 2*x1*x2^3 + x1^2*x2*x3
                sage: y = W.from_reduced_word([3,1,2,0,3,1,0])
                sage: y.stanley_symmetric_function_as_polynomial()  # long time
                8*x1^7 + 4*x1^5*x2 + 2*x1^3*x2^2 + x1*x2^3

                sage: W = WeylGroup(['C',3,1])
                sage: W.from_reduced_word([0,2,1,0]).stanley_symmetric_function_as_polynomial()
                32*x1^4 + 16*x1^2*x2 + 8*x2^2 + 4*x1*x3

                sage: W = WeylGroup(['B',3,1])
                sage: W.from_reduced_word([3,2,1]).stanley_symmetric_function_as_polynomial()
                2*x1^3 + x1*x2 + 1/2*x3

            Algorithm: Induction on the left Pieri factors. Note that
            this induction preserves subsets of `W` which are stable
            by taking right factors, and in particular Grassmanian
            elements.
            """
            if max_length is None:
                from sage.rings.infinity import infinity
                max_length = infinity
            W = self.parent()
            pieri_factors = W.pieri_factors()
            from sage.rings.rational_field import QQ
            R = QQ[','.join('x%s' % l for l in range(1, pieri_factors.max_length() + 1))]
            x = R.gens()
            if self.is_one():
                return R.one()

            return R(sum(2**(pieri_factors.stanley_symm_poly_weight(u)) * x[u.length() - 1] * v.stanley_symmetric_function_as_polynomial(max_length=u.length())
                         for (u, v) in self.left_pieri_factorizations(max_length)
                         if u != W.one()))

        def stanley_symmetric_function(self):
            r"""
            Return the affine Stanley symmetric function indexed by ``self``.

            INPUT:

            - ``self`` -- an element `w` of a Weyl group

            Returns the affine Stanley symmetric function indexed by
            `w`. Stanley symmetric functions are defined as generating
            series of the factorizations of `w` into Pieri factors and
            weighted by a statistic on Pieri factors.

            .. SEEALSO::

                * :meth:`stanley_symmetric_function_as_polynomial`
                * :meth:`WeylGroups.ParentMethods.pieri_factors`
                * :mod:`sage.combinat.root_system.pieri_factors`

            EXAMPLES::

                sage: W = WeylGroup(['A', 3, 1])
                sage: W.from_reduced_word([3,1,2,0,3,1,0]).stanley_symmetric_function()
                8*m[1, 1, 1, 1, 1, 1, 1] + 4*m[2, 1, 1, 1, 1, 1]
                + 2*m[2, 2, 1, 1, 1] + m[2, 2, 2, 1]
                sage: A = AffinePermutationGroup(['A',3,1])
                sage: A.from_reduced_word([3,1,2,0,3,1,0]).stanley_symmetric_function()
                8*m[1, 1, 1, 1, 1, 1, 1] + 4*m[2, 1, 1, 1, 1, 1]
                + 2*m[2, 2, 1, 1, 1] + m[2, 2, 2, 1]

                sage: W = WeylGroup(['C',3,1])
                sage: W.from_reduced_word([0,2,1,0]).stanley_symmetric_function()
                32*m[1, 1, 1, 1] + 16*m[2, 1, 1] + 8*m[2, 2] + 4*m[3, 1]

                sage: W = WeylGroup(['B',3,1])
                sage: W.from_reduced_word([3,2,1]).stanley_symmetric_function()
                2*m[1, 1, 1] + m[2, 1] + 1/2*m[3]

                sage: W = WeylGroup(['B',4])
                sage: w = W.from_reduced_word([3,2,3,1])
                sage: w.stanley_symmetric_function()  # long time (6s on sage.math, 2011)
                48*m[1, 1, 1, 1] + 24*m[2, 1, 1] + 12*m[2, 2] + 8*m[3, 1] + 2*m[4]

                sage: A = AffinePermutationGroup(['A',4,1])
                sage: a = A([-2,0,1,4,12])
                sage: a.stanley_symmetric_function()
                6*m[1, 1, 1, 1, 1, 1, 1, 1] + 5*m[2, 1, 1, 1, 1, 1, 1]
                + 4*m[2, 2, 1, 1, 1, 1] + 3*m[2, 2, 2, 1, 1] + 2*m[2, 2, 2, 2]
                + 4*m[3, 1, 1, 1, 1, 1] + 3*m[3, 2, 1, 1, 1] + 2*m[3, 2, 2, 1]
                + 2*m[3, 3, 1, 1] + m[3, 3, 2] + 3*m[4, 1, 1, 1, 1]
                + 2*m[4, 2, 1, 1] + m[4, 2, 2] + m[4, 3, 1]

            One more example (:issue:`14095`)::

                sage: G = SymmetricGroup(4)
                sage: w = G.from_reduced_word([3,2,3,1])
                sage: w.stanley_symmetric_function()
                3*m[1, 1, 1, 1] + 2*m[2, 1, 1] + m[2, 2] + m[3, 1]

            REFERENCES:

            - [BH1994]_

            - [Lam2008]_

            - [LSS2009]_

            - [Pon2010]_
            """
            from sage.rings.rational_field import QQ
            from sage.combinat.sf.sf import SymmetricFunctions

            m = SymmetricFunctions(QQ).monomial()
            return m.from_polynomial_exp(self.stanley_symmetric_function_as_polynomial())

        @cached_in_parent_method
        def reflection_to_root(self):
            r"""
            Return the root associated with the reflection ``self``.

            EXAMPLES::

                sage: W = WeylGroup(['C',2], prefix='s')
                sage: W.from_reduced_word([1,2,1]).reflection_to_root()
                2*alpha[1] + alpha[2]
                sage: W.from_reduced_word([1,2]).reflection_to_root()
                Traceback (most recent call last):
                ...
                ValueError: s1*s2 is not a reflection
                sage: W.long_element().reflection_to_root()
                Traceback (most recent call last):
                ...
                ValueError: s2*s1*s2*s1 is not a reflection
            """
            i = self.first_descent()
            if i is None:
                raise ValueError("{} is not a reflection".format(self))
            if self == self.parent().simple_reflection(i):
                return self.parent().cartan_type().root_system().root_lattice().simple_root(i)
            rsi = self.apply_simple_reflection(i)
            if not rsi.has_descent(i, side='left'):
                raise ValueError("{} is not a reflection".format(self))
            return rsi.apply_simple_reflection(i, side='left').reflection_to_root().simple_reflection(i)

        @cached_in_parent_method
        def reflection_to_coroot(self):
            r"""
            Return the coroot associated with the reflection ``self``.

            EXAMPLES::

                sage: W = WeylGroup(['C',2], prefix='s')
                sage: W.from_reduced_word([1,2,1]).reflection_to_coroot()
                alphacheck[1] + alphacheck[2]
                sage: W.from_reduced_word([1,2]).reflection_to_coroot()
                Traceback (most recent call last):
                ...
                ValueError: s1*s2 is not a reflection
                sage: W.long_element().reflection_to_coroot()
                Traceback (most recent call last):
                ...
                ValueError: s2*s1*s2*s1 is not a reflection
            """
            i = self.first_descent()
            if i is None:
                raise ValueError("{} is not a reflection".format(self))
            if self == self.parent().simple_reflection(i):
                return self.parent().cartan_type().root_system().root_lattice().simple_coroot(i)
            rsi = self.apply_simple_reflection(i)
            if not rsi.has_descent(i, side='left'):
                raise ValueError("{} is not a reflection".format(self))
            return rsi.apply_simple_reflection(i, side='left').reflection_to_coroot().simple_reflection(i)

        def inversions(self, side='right', inversion_type='reflections'):
            """
            Return the set of inversions of ``self``.

            INPUT:

            - ``side`` -- 'right' (default) or 'left'
            - ``inversion_type`` -- 'reflections' (default), 'roots', or 'coroots'

            OUTPUT:

            For reflections, the set of reflections r in the Weyl group such that
            ``self`` ``r`` < ``self``. For (co)roots, the set of positive (co)roots that are sent
            by ``self`` to negative (co)roots; their associated reflections are described above.

            If ``side`` is 'left', the inverse Weyl group element is used.

            EXAMPLES::

                sage: W = WeylGroup(['C',2], prefix='s')
                sage: w = W.from_reduced_word([1,2])
                sage: w.inversions()
                [s2, s2*s1*s2]
                sage: w.inversions(inversion_type = 'reflections')
                [s2, s2*s1*s2]
                sage: w.inversions(inversion_type = 'roots')
                [alpha[2], alpha[1] + alpha[2]]
                sage: w.inversions(inversion_type = 'coroots')
                [alphacheck[2], alphacheck[1] + 2*alphacheck[2]]
                sage: w.inversions(side = 'left')
                [s1, s1*s2*s1]
                sage: w.inversions(side = 'left', inversion_type = 'roots')
                [alpha[1], 2*alpha[1] + alpha[2]]
                sage: w.inversions(side = 'left', inversion_type = 'coroots')
                [alphacheck[1], alphacheck[1] + alphacheck[2]]
            """
            if side == 'left':
                self = self.inverse()
            reflections = self.inversions_as_reflections()
            if inversion_type == 'reflections':
                return reflections
            if inversion_type == 'roots':
                return [r.reflection_to_root() for r in reflections]
            if inversion_type == 'coroots':
                return [r.reflection_to_coroot() for r in reflections]
            raise ValueError(f"inversion_type {inversion_type} is invalid")

        def inversion_arrangement(self, side='right'):
            r"""
            Return the inversion hyperplane arrangement of ``self``.

            INPUT:

            - ``side`` -- ``'right'`` (default) or ``'left'``

            OUTPUT:

            A (central) hyperplane arrangement whose hyperplanes correspond
            to the inversions of ``self`` given as roots.

            The ``side`` parameter determines on which side
            to compute the inversions.

            EXAMPLES::

                sage: W = WeylGroup(['A',3])
                sage: w = W.from_reduced_word([1, 2, 3, 1, 2])
                sage: A = w.inversion_arrangement(); A
                Arrangement of 5 hyperplanes of dimension 3 and rank 3
                sage: A.hyperplanes()
                (Hyperplane 0*a1 + 0*a2 + a3 + 0,
                 Hyperplane 0*a1 + a2 + 0*a3 + 0,
                 Hyperplane 0*a1 + a2 + a3 + 0,
                 Hyperplane a1 + a2 + 0*a3 + 0,
                 Hyperplane a1 + a2 + a3 + 0)

            The identity element gives the empty arrangement::

                sage: W = WeylGroup(['A',3])
                sage: W.one().inversion_arrangement()
                Empty hyperplane arrangement of dimension 3
            """
            inv = self.inversions(side=side, inversion_type='roots')
            from sage.geometry.hyperplane_arrangement.arrangement import HyperplaneArrangements
            I = self.parent().cartan_type().index_set()
            from sage.rings.rational_field import QQ
            H = HyperplaneArrangements(QQ, tuple(['a{}'.format(i) for i in I]))
            gens = H.gens()
            if not inv:
                return H()
            return H([sum(c * gens[I.index(i)] for (i, c) in root)
                      for root in inv])

        def bruhat_lower_covers_coroots(self):
            r"""
            Return all 2-tuples (``v``, `\alpha`) where ``v`` is covered
            by ``self`` and `\alpha` is the positive coroot such that
            ``self`` = ``v`` `s_\alpha` where `s_\alpha` is
            the reflection orthogonal to `\alpha`.

            ALGORITHM:

            See :meth:`.bruhat_lower_covers` and
            :meth:`.bruhat_lower_covers_reflections` for Coxeter groups.

            EXAMPLES::

                sage: W = WeylGroup(['A',3], prefix='s')
                sage: w = W.from_reduced_word([3,1,2,1])
                sage: w.bruhat_lower_covers_coroots()
                [(s1*s2*s1, alphacheck[1] + alphacheck[2] + alphacheck[3]),
                 (s3*s2*s1, alphacheck[2]), (s3*s1*s2, alphacheck[1])]
            """
            return [(x[0], x[1].reflection_to_coroot())
                    for x in self.bruhat_lower_covers_reflections()]

        def bruhat_upper_covers_coroots(self):
            r"""
            Return all 2-tuples (``v``, `\alpha`) where ``v`` is covers ``self`` and `\alpha`
            is the positive coroot such that ``self`` = ``v`` `s_\alpha` where `s_\alpha` is
            the reflection orthogonal to `\alpha`.

            ALGORITHM:

            See :meth:`~CoxeterGroups.ElementMethods.bruhat_upper_covers` and :meth:`.bruhat_upper_covers_reflections` for Coxeter groups.

            EXAMPLES::

                sage: W = WeylGroup(['A',4], prefix='s')
                sage: w = W.from_reduced_word([3,1,2,1])
                sage: w.bruhat_upper_covers_coroots()
                [(s1*s2*s3*s2*s1, alphacheck[3]),
                 (s2*s3*s1*s2*s1, alphacheck[2] + alphacheck[3]),
                 (s3*s4*s1*s2*s1, alphacheck[4]),
                 (s4*s3*s1*s2*s1, alphacheck[1] + alphacheck[2] + alphacheck[3] + alphacheck[4])]
            """
            return [(x[0], x[1].reflection_to_coroot())
                    for x in self.bruhat_upper_covers_reflections()]

        def quantum_bruhat_successors(self, index_set=None, roots=False, quantum_only=False):
            r"""
            Return the successors of ``self`` in the quantum Bruhat graph
            on the parabolic quotient of the Weyl group determined by the
            subset of Dynkin nodes ``index_set``.

            INPUT:

            - ``self`` -- a Weyl group element, which is assumed to
              be of minimum length in its coset with respect to the
              parabolic subgroup

            - ``index_set`` -- (default: ``None``) indicates the set of
              simple reflections used to generate the parabolic subgroup;
              the default value indicates that the subgroup is the identity

            - ``roots`` -- boolean (default: ``False``); if ``True``, returns
              the list of 2-tuples (``w``, `\alpha`) where ``w`` is a successor
              and `\alpha` is the positive root associated with the
              successor relation

            - ``quantum_only`` -- boolean (default: ``False``); if ``True``,
              returns only the quantum successors

            EXAMPLES::

                sage: W = WeylGroup(['A',3], prefix='s')
                sage: w = W.from_reduced_word([3,1,2])
                sage: w.quantum_bruhat_successors([1], roots = True)
                [(s3, alpha[2]), (s1*s2*s3*s2, alpha[3]),
                 (s2*s3*s1*s2, alpha[1] + alpha[2] + alpha[3])]
                sage: w.quantum_bruhat_successors([1,3])
                [1, s2*s3*s1*s2]
                sage: w.quantum_bruhat_successors(roots = True)
                [(s3*s1*s2*s1, alpha[1]),
                 (s3*s1, alpha[2]),
                 (s1*s2*s3*s2, alpha[3]),
                 (s2*s3*s1*s2, alpha[1] + alpha[2] + alpha[3])]
                sage: w.quantum_bruhat_successors()
                [s3*s1*s2*s1, s3*s1, s1*s2*s3*s2, s2*s3*s1*s2]
                sage: w.quantum_bruhat_successors(quantum_only = True)
                [s3*s1]
                sage: w = W.from_reduced_word([2,3])
                sage: w.quantum_bruhat_successors([1,3])
                Traceback (most recent call last):
                ...
                ValueError: s2*s3 is not of minimum length in its coset
                of the parabolic subgroup generated by the reflections (1, 3)
            """
            W = self.parent()
            if not W.cartan_type().is_finite():
                raise ValueError("the Cartan type {} is not finite".format(W.cartan_type()))
            if index_set is None:
                index_set = []
            else:
                index_set = list(index_set)
            index_set = tuple(index_set)
            if self != self.coset_representative(index_set):
                raise ValueError("{} is not of minimum length in its coset of the parabolic subgroup generated by the reflections {}".format(self, index_set))
            lattice = W.cartan_type().root_system().root_lattice()
            w_length_plus_one = self.length() + 1
            successors = []
            for alpha in lattice.nonparabolic_positive_roots(index_set):
                wr = self * W.from_reduced_word(alpha.associated_reflection())
                wrc = wr.coset_representative(index_set)
                if wrc == wr and wr.length() == w_length_plus_one and not quantum_only:
                    if roots:
                        successors.append((wr, alpha))
                    else:
                        successors.append(wr)
                elif alpha.quantum_root() and wrc.length() == w_length_plus_one - lattice.nonparabolic_positive_root_sum(index_set).scalar(alpha.associated_coroot()):
                    if roots:
                        successors.append((wrc, alpha))
                    else:
                        successors.append(wrc)
            return successors



# sage_setup: distribution = sagemath-categories
# sage.doctest: needs sage.libs.pari
"""
Examples of sets
"""
# ****************************************************************************
#  Copyright (C) 2009 Florent Hivert <Florent.Hivert@univ-rouen.fr>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.structure.parent import Parent
from sage.structure.element import Element
from sage.categories.sets_cat import Sets
from sage.rings.integer import Integer, IntegerWrapper
from sage.rings.integer_ring import IntegerRing
from sage.arith.misc import is_prime
from sage.structure.unique_representation import UniqueRepresentation


class PrimeNumbers(UniqueRepresentation, Parent):
    r"""
    An example of parent in the category of sets: the set of prime numbers.

    The elements are represented as plain integers in `\ZZ` (facade
    implementation).

    This is a minimal implementations. For more advanced examples of
    implementations, see also::

        sage: P = Sets().example("facade")
        sage: P = Sets().example("inherits")
        sage: P = Sets().example("wrapper")

    EXAMPLES::

        sage: P = Sets().example()
        sage: P(12)
        Traceback (most recent call last):
        ...
        AssertionError: 12 is not a prime number
        sage: a = P.an_element()
        sage: a.parent()
        Integer Ring
        sage: x = P(13); x
        13
        sage: type(x)
        <class 'sage.rings.integer.Integer'>
        sage: x.parent()
        Integer Ring
        sage: 13 in P
        True
        sage: 12 in P
        False
        sage: y = x+1; y
        14
        sage: type(y)
        <class 'sage.rings.integer.Integer'>

        sage: TestSuite(P).run(verbose=True)
        running ._test_an_element() . . . pass
        running ._test_cardinality() . . . pass
        running ._test_category() . . . pass
        running ._test_construction() . . . pass
        running ._test_elements() . . .
          Running the test suite of self.an_element()
          running ._test_category() . . . pass
          running ._test_eq() . . . pass
          running ._test_new() . . . pass
          running ._test_nonzero_equal() . . . pass
          running ._test_not_implemented_methods() . . . pass
          running ._test_pickling() . . . pass
          pass
        running ._test_elements_eq_reflexive() . . . pass
        running ._test_elements_eq_symmetric() . . . pass
        running ._test_elements_eq_transitive() . . . pass
        running ._test_elements_neq() . . . pass
        running ._test_eq() . . . pass
        running ._test_new() . . . pass
        running ._test_not_implemented_methods() . . . pass
        running ._test_pickling() . . . pass
        running ._test_some_elements() . . . pass
    """
    def __init__(self):
        """
        TESTS::

            sage: from sage.categories.examples.sets_cat import PrimeNumbers
            sage: P = PrimeNumbers()
            sage: P.category()
            Category of facade sets
            sage: P is Sets().example()
            True
        """
        Parent.__init__(self, facade=IntegerRing(), category=Sets())

    def _repr_(self):
        """
        TESTS::

            sage: Sets().example()       # indirect doctest
            Set of prime numbers (basic implementation)
        """
        return "Set of prime numbers (basic implementation)"

    def an_element(self):
        """
        Implement :meth:`Sets.ParentMethods.an_element`.

        TESTS::

            sage: P = Sets().example()
            sage: x = P.an_element(); x
            47
            sage: x.parent()
            Integer Ring
        """
        return self(47) # if speed is needed, call: self.element_class(47)

    def __contains__(self, p):
        """
        TESTS::

            sage: P = Sets().example()
            sage: 13 in P
            True
            sage: 12 in P
            False
        """
        return isinstance(p, Integer) and p.is_prime()

    def _element_constructor_(self, e):
        """
        TESTS::

            sage: P = Sets().example()
            sage: P._element_constructor_(13) == 13
            True
            sage: P._element_constructor_(13).parent()
            Integer Ring
            sage: P._element_constructor_(14)
            Traceback (most recent call last):
            ...
            AssertionError: 14 is not a prime number
        """
        p = self.element_class(e)
        assert is_prime(p), "%s is not a prime number" % (p)
        return p

    element_class = Integer


from sage.misc.abstract_method import abstract_method


class PrimeNumbers_Abstract(UniqueRepresentation, Parent):
    """
    This class shows how to write a parent while keeping the choice of the
    datastructure for the children open. Different class with fixed
    datastructure will then be constructed by inheriting from
    :class:`PrimeNumbers_Abstract`.

    This is used by::

        sage: P = Sets().example("facade")
        sage: P = Sets().example("inherits")
        sage: P = Sets().example("wrapper")
    """
    def __init__(self):
        """
        TESTS::

            sage: P = Sets().example("inherits")
        """
        Parent.__init__(self, category=Sets())

    def _repr_(self):
        """
        TESTS::

            sage: Sets().example("inherits")     # indirect doctest
            Set of prime numbers
        """
        return "Set of prime numbers"

    def an_element(self):
        """
        Implement :meth:`Sets.ParentMethods.an_element`.

        TESTS::

            sage: P = Sets().example("inherits")
            sage: x = P.an_element(); x
            47
            sage: x.parent()
            Set of prime numbers
        """
        return self._from_integer_(47)

    def _element_constructor_(self, i):
        """
        Construct an element of ``self`` from an integer, testing that
        this integer is indeed prime.

        EXAMPLES::

            sage: P = Sets().example("inherits")
            sage: P(13)       # indirect doctest
            13
            sage: P(42)
            Traceback (most recent call last):
            ...
            ValueError: 42 is not a prime number
        """
        if i in self:
            return self._from_integer_(i)
        else:
            raise ValueError("%s is not a prime number" % (i))

    @abstract_method
    def _from_integer_(self, i):
        """
        Fast construction of an element of ``self`` from an integer.

        No prime checking is performed. To be defined.

        EXAMPLES::

            sage: P = Sets().example("inherits")
            sage: P._from_integer_(13)
            13
            sage: P._from_integer_(42)       # Do not do that at home kids!
            42
            sage: P(42)
            Traceback (most recent call last):
            ...
            ValueError: 42 is not a prime number
        """

    def next(self, i):
        """
        Return the next prime number.

        EXAMPLES::

            sage: P = Sets().example("inherits")
            sage: x = P.next(P.an_element()); x
            53
            sage: x.parent()
            Set of prime numbers
        """
        assert i in self
        return self._from_integer_((Integer(i) + 1).next_prime())

    def some_elements(self):
        """
        Return some prime numbers.

        EXAMPLES::

            sage: P = Sets().example("inherits")
            sage: P.some_elements()
            [47, 53, 59, 61, 67, 71, 73, 79, 83, 89, 97]
        """
        x = self.an_element()
        res = [x]
        for i in range(10):
            x = self.next(x)
            res.append(x)
        return res

    class Element(Element):
        def is_prime(self):
            """
            Return whether ``self`` is a prime number.

            EXAMPLES::

                sage: P = Sets().example("inherits")
                sage: x = P.an_element()
                sage: P.an_element().is_prime()
                True
            """
            return True

        def next(self):
            """
            Return the next prime number.

            EXAMPLES::

                sage: P = Sets().example("inherits")
                sage: p = P.an_element(); p
                47
                sage: p.next()
                53

            .. NOTE::

                This method is not meant to implement the protocol iterator,
                and thus not subject to Python 2 vs Python 3 incompatibilities.
            """
            return self.parent().next(self)


#*************************************************************************#
class PrimeNumbers_Inherits(PrimeNumbers_Abstract):
    """
    An example of parent in the category of sets: the set of prime numbers.
    In this implementation, the element are stored as object of a new class
    which inherits from the class Integer (technically :class:`IntegerWrapper`).

    EXAMPLES::

        sage: P = Sets().example("inherits")
        sage: P
        Set of prime numbers
        sage: P(12)
        Traceback (most recent call last):
        ...
        ValueError: 12 is not a prime number
        sage: a = P.an_element()
        sage: a.parent()
        Set of prime numbers
        sage: x = P(13); x
        13
        sage: x.is_prime()
        True
        sage: type(x)
        <class 'sage.categories.examples.sets_cat.PrimeNumbers_Inherits_with_category.element_class'>
        sage: x.parent()
        Set of prime numbers
        sage: P(13) in P
        True
        sage: y = x+1; y
        14
        sage: type(y)
        <class 'sage.rings.integer.Integer'>
        sage: y.parent()
        Integer Ring
        sage: type(P(13)+P(17))
        <class 'sage.rings.integer.Integer'>
        sage: type(P(2)+P(3))
        <class 'sage.rings.integer.Integer'>

        sage: z = P.next(x); z
        17
        sage: type(z)
        <class 'sage.categories.examples.sets_cat.PrimeNumbers_Inherits_with_category.element_class'>
        sage: z.parent()
        Set of prime numbers

        sage: TestSuite(P).run(verbose=True)
        running ._test_an_element() . . . pass
        running ._test_cardinality() . . . pass
        running ._test_category() . . . pass
        running ._test_construction() . . . pass
        running ._test_elements() . . .
          Running the test suite of self.an_element()
          running ._test_category() . . . pass
          running ._test_eq() . . . pass
          running ._test_new() . . . pass
          running ._test_not_implemented_methods() . . . pass
          running ._test_pickling() . . . pass
          pass
        running ._test_elements_eq_reflexive() . . . pass
        running ._test_elements_eq_symmetric() . . . pass
        running ._test_elements_eq_transitive() . . . pass
        running ._test_elements_neq() . . . pass
        running ._test_eq() . . . pass
        running ._test_new() . . . pass
        running ._test_not_implemented_methods() . . . pass
        running ._test_pickling() . . . pass
        running ._test_some_elements() . . . pass

    See also::

        sage: P = Sets().example("facade")
        sage: P = Sets().example("inherits")
        sage: P = Sets().example("wrapper")
    """

    def __init__(self):
        """
        TESTS::

            sage: P = Sets().example("inherits")
            sage: type(P(13)+P(17))
            <class 'sage.rings.integer.Integer'>
            sage: type(P(2)+P(3))
            <class 'sage.rings.integer.Integer'>
        """
        super().__init__()
        self._populate_coercion_lists_(embedding=IntegerRing())

    def __contains__(self, p):
        """
        TESTS::

            sage: P = Sets().example("inherits")
            sage: 13 in P, P(13) in P
            (True, True)
            sage: 12 in P
            False
        """
        return (isinstance(p, self.element_class) and p.parent() is self
                or isinstance(p, Integer) and p.is_prime())

    def _from_integer_(self, p):
        """
        TESTS::

            sage: P = Sets().example("inherits")
            sage: P._from_integer_(13)
            13
            sage: P._from_integer_(42)            # Don't do that at home kids!
            42
        """
        return self.element_class(self, p)

    class Element(IntegerWrapper, PrimeNumbers_Abstract.Element):
        def __init__(self, parent, p):
            """
            TESTS::

                sage: P = Sets().example("inherits")
                sage: P(12)
                Traceback (most recent call last):
                ...
                ValueError: 12 is not a prime number
                sage: x = P(13); type(x)
                <class 'sage.categories.examples.sets_cat.PrimeNumbers_Inherits_with_category.element_class'>
                sage: x.parent() is P
                True
            """
            IntegerWrapper.__init__(self, parent, p)


#*************************************************************************#
class PrimeNumbers_Wrapper(PrimeNumbers_Abstract):
    """
    An example of parent in the category of sets: the set of prime numbers.

    In this second alternative implementation, the prime integer are stored as
    a attribute of a sage object by inheriting from :class:`ElementWrapper`.  In
    this case we need to ensure conversion and coercion from this parent and
    its element to ``ZZ`` and ``Integer``.

    EXAMPLES::

        sage: P = Sets().example("wrapper")
        sage: P(12)
        Traceback (most recent call last):
        ...
        ValueError: 12 is not a prime number
        sage: a = P.an_element()
        sage: a.parent()
        Set of prime numbers (wrapper implementation)
        sage: x = P(13); x
        13
        sage: type(x)
        <class 'sage.categories.examples.sets_cat.PrimeNumbers_Wrapper_with_category.element_class'>
        sage: x.parent()
        Set of prime numbers (wrapper implementation)
        sage: 13 in P
        True
        sage: 12 in P
        False
        sage: y = x+1; y
        14
        sage: type(y)
        <class 'sage.rings.integer.Integer'>

        sage: z = P.next(x); z
        17
        sage: type(z)
        <class 'sage.categories.examples.sets_cat.PrimeNumbers_Wrapper_with_category.element_class'>
        sage: z.parent()
        Set of prime numbers (wrapper implementation)

    TESTS::

        sage: TestSuite(P).run()
    """
    def __init__(self):
        """
        TESTS::

            sage: P = Sets().example("wrapper")
            sage: P.category()
            Category of sets
            sage: P(13) == 13
            True
            sage: ZZ(P(13)) == 13
            True
            sage: P(13) + 1 == 14
            True
        """
        Parent.__init__(self, category=Sets())
        from sage.rings.integer_ring import IntegerRing
        from sage.categories.homset import Hom
        self.mor = Hom(self, IntegerRing())(lambda z: z.value)
        self._populate_coercion_lists_(embedding=self.mor)

    def _repr_(self):
        """
        TESTS::

            sage: Sets().example("wrapper")       # indirect doctest
            Set of prime numbers (wrapper implementation)
        """
        return "Set of prime numbers (wrapper implementation)"

    def __contains__(self, p):
        """
        TESTS::

            sage: P = Sets().example("wrapper")
            sage: 13 in P
            True
            sage: 12 in P
            False
        """
        return (isinstance(p, self.element_class) and p.parent() == self or
                isinstance(p, Integer) and p.is_prime())

    def _from_integer_(self, e):
        """
        TESTS::

            sage: P = Sets().example("wrapper")
            sage: P._from_integer_(13).parent()
            Set of prime numbers (wrapper implementation)
            sage: P._from_integer_(14)            # Don't do that at home kids!
            14
            sage: P._element_constructor_(14)
            Traceback (most recent call last):
            ...
            ValueError: 14 is not a prime number
        """
        return self.element_class(self, Integer(e))

    from sage.structure.element_wrapper import ElementWrapper

    class Element (ElementWrapper, PrimeNumbers_Abstract.Element):
        def _integer_(self, IntRing):
            """
            Convert to an integer.

            TESTS::

                sage: P = Sets().example("wrapper")
                sage: x = P.an_element()
                sage: Integer(x)           # indirect doctest
                47
            """
            return IntRing(self.value)


#*************************************************************************#
class PrimeNumbers_Facade(PrimeNumbers_Abstract):
    r"""
    An example of parent in the category of sets: the set of prime numbers.

    In this alternative implementation, the elements are represented
    as plain integers in `\ZZ` (facade implementation).

    EXAMPLES::

        sage: P = Sets().example("facade")
        sage: P(12)
        Traceback (most recent call last):
        ...
        ValueError: 12 is not a prime number
        sage: a = P.an_element()
        sage: a.parent()
        Integer Ring
        sage: x = P(13); x
        13
        sage: type(x)
        <class 'sage.rings.integer.Integer'>
        sage: x.parent()
        Integer Ring
        sage: 13 in P
        True
        sage: 12 in P
        False
        sage: y = x+1; y
        14
        sage: type(y)
        <class 'sage.rings.integer.Integer'>

        sage: z = P.next(x); z
        17
        sage: type(z)
        <class 'sage.rings.integer.Integer'>
        sage: z.parent()
        Integer Ring

    The disadvantage of this implementation is that the elements do not know
    that they are prime, so that prime testing is slow::

        sage: pf = Sets().example("facade").an_element()
        sage: timeit("pf.is_prime()") #    random
        625 loops, best of 3: 4.1 us per loop

    compared to the other implementations where prime testing is only done if
    needed during the construction of the element, and later on the elements
    "know" that they are prime::

        sage: pw = Sets().example("wrapper").an_element()
        sage: timeit("pw.is_prime()")    # random
        625 loops, best of 3: 859 ns per loop

        sage: pi = Sets().example("inherits").an_element()
        sage: timeit("pw.is_prime()")    # random
        625 loops, best of 3: 854 ns per loop

    Note also that the ``next`` method for the elements does not exist::

        sage: pf.next()
        Traceback (most recent call last):
        ...
        AttributeError: 'sage.rings.integer.Integer' object has no attribute 'next'...

    unlike in the other implementations::

        sage: pw.next()
        53
        sage: pi.next()
        53

    TESTS::

        sage: TestSuite(P).run(verbose = True)
        running ._test_an_element() . . . pass
        running ._test_cardinality() . . . pass
        running ._test_category() . . . pass
        running ._test_construction() . . . pass
        running ._test_elements() . . .
          Running the test suite of self.an_element()
          running ._test_category() . . . pass
          running ._test_eq() . . . pass
          running ._test_new() . . . pass
          running ._test_nonzero_equal() . . . pass
          running ._test_not_implemented_methods() . . . pass
          running ._test_pickling() . . . pass
          pass
        running ._test_elements_eq_reflexive() . . . pass
        running ._test_elements_eq_symmetric() . . . pass
        running ._test_elements_eq_transitive() . . . pass
        running ._test_elements_neq() . . . pass
        running ._test_eq() . . . pass
        running ._test_new() . . . pass
        running ._test_not_implemented_methods() . . . pass
        running ._test_pickling() . . . pass
        running ._test_some_elements() . . . pass
    """

    def __init__(self):
        """
        TESTS::

            sage: P = Sets().example("inherits")
        """
        Parent.__init__(self, facade=IntegerRing(), category=Sets())

    def _repr_(self):
        """
        TESTS::

            sage: Sets().example("facade")       # indirect doctest
            Set of prime numbers (facade implementation)
        """
        return "Set of prime numbers (facade implementation)"

    def __contains__(self, p):
        """
        TESTS::

            sage: P = Sets().example("facade")
            sage: 13 in P
            True
            sage: 12 in P
            False
        """
        return isinstance(p, Integer) and p.is_prime()

    def _from_integer_(self, e):
        """
        TESTS::

            sage: P = Sets().example("facade")
            sage: P._from_integer_(13).parent()
            Integer Ring
            sage: P._from_integer_(14)            # Don't do that at home kids!
            14
            sage: P._element_constructor_(14)
            Traceback (most recent call last):
            ...
            ValueError: 14 is not a prime number
        """
        return self.element_class(e)

    element_class = Integer



# sage_setup: distribution = sagemath-categories
# sage.doctest: needs sage.combinat sage.graphs
r"""
Regular Crystals
"""
#*****************************************************************************
#  Copyright (C) 2013    Anne Schilling <anne at math.ucdavis.edu>
#                        Travis Scrimshaw <tscrim at ucdavis.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#****************************************************************************

from sage.misc.cachefunc import cached_method
from sage.categories.category_singleton import Category_singleton
from sage.categories.crystals import Crystals
from sage.categories.tensor import TensorProductsCategory


class RegularCrystals(Category_singleton):
    r"""
    The category of regular crystals.

    A crystal is called *regular* if every vertex `b` satisfies

    .. MATH::

        \varepsilon_i(b) = \max\{ k \mid e_i^k(b) \neq 0 \} \quad \text{and}
        \quad \varphi_i(b) = \max\{ k \mid f_i^k(b) \neq 0 \}.

    .. NOTE::

        Regular crystals are sometimes referred to as *normal*. When only one
        of the conditions (on either `\varphi_i` or `\varepsilon_i`) holds,
        these crystals are sometimes called *seminormal* or *semiregular*.

    EXAMPLES::

        sage: C = RegularCrystals()
        sage: C
        Category of regular crystals
        sage: C.super_categories()
        [Category of crystals]
        sage: C.example()
        Highest weight crystal of type A_3 of highest weight omega_1

    TESTS::

        sage: TestSuite(C).run()
        sage: B = RegularCrystals().example()
        sage: TestSuite(B).run(verbose = True)
        running ._test_an_element() . . . pass
        running ._test_cardinality() . . . pass
        running ._test_category() . . . pass
        running ._test_construction() . . . pass
        running ._test_elements() . . .
          Running the test suite of self.an_element()
          running ._test_category() . . . pass
          running ._test_eq() . . . pass
          running ._test_new() . . . pass
          running ._test_not_implemented_methods() . . . pass
          running ._test_pickling() . . . pass
          running ._test_stembridge_local_axioms() . . . pass
          pass
        running ._test_elements_eq_reflexive() . . . pass
        running ._test_elements_eq_symmetric() . . . pass
        running ._test_elements_eq_transitive() . . . pass
        running ._test_elements_neq() . . . pass
        running ._test_enumerated_set_contains() . . . pass
        running ._test_enumerated_set_iter_cardinality() . . . pass
        running ._test_enumerated_set_iter_list() . . . pass
        running ._test_eq() . . . pass
        running ._test_fast_iter() . . . pass
        running ._test_new() . . . pass
        running ._test_not_implemented_methods() . . . pass
        running ._test_pickling() . . . pass
        running ._test_some_elements() . . . pass
        running ._test_stembridge_local_axioms() . . . pass
    """

    @cached_method
    def super_categories(self):
        r"""
        EXAMPLES::

            sage: RegularCrystals().super_categories()
            [Category of crystals]
        """
        return [Crystals()]

    def example(self, n=3):
        """
        Return an example of highest weight crystals, as per
        :meth:`Category.example`.

        EXAMPLES::

            sage: B = RegularCrystals().example(); B
            Highest weight crystal of type A_3 of highest weight omega_1
        """
        from sage.categories.crystals import Crystals
        return Crystals().example(n)

    def additional_structure(self):
        r"""
        Return ``None``.

        Indeed, the category of regular crystals defines no new
        structure: it only relates `\varepsilon_a` and `\varphi_a` to
        `e_a` and `f_a` respectively.

        .. SEEALSO:: :meth:`Category.additional_structure`

        .. TODO:: Should this category be a :class:`CategoryWithAxiom`?

        EXAMPLES::

            sage: RegularCrystals().additional_structure()
        """
        return None

    class MorphismMethods:
        def is_isomorphism(self):
            """
            Check if ``self`` is a crystal isomorphism, which is true
            if and only if this is a strict embedding with the same number
            of connected components.

            EXAMPLES::

                sage: A21 = RootSystem(['A',2,1])
                sage: La = A21.weight_space(extended=True).fundamental_weights()
                sage: B = crystals.LSPaths(La[0])
                sage: La = A21.weight_lattice(extended=True).fundamental_weights()
                sage: C = crystals.GeneralizedYoungWalls(2, La[0])
                sage: H = Hom(B, C)
                sage: from sage.categories.highest_weight_crystals import HighestWeightCrystalMorphism
                sage: class Psi(HighestWeightCrystalMorphism):
                ....:     def is_strict(self):
                ....:         return True
                sage: psi = Psi(H, C.module_generators); psi
                ['A', 2, 1] Crystal morphism:
                  From: The crystal of LS paths of type ['A', 2, 1] and weight Lambda[0]
                  To:   Highest weight crystal of generalized Young walls
                        of Cartan type ['A', 2, 1] and highest weight Lambda[0]
                  Defn: (Lambda[0],) |--> []
                sage: psi.is_isomorphism()
                True
            """
            return (self.is_strict()
                    and self.domain().number_of_connected_components() ==
                        self.codomain().number_of_connected_components())

    class ParentMethods:

        # TODO: this could be a method in Crystals.Algebras.ElementMethods, so that
        # one could do:
        #
        #   sage: C = crystals.Tableaux(['A',2], shape=[2,1])
        #   sage: M = C.algebra(QQ)
        #   sage: m = M.an_element()
        #   sage: m.demazure_operator([1,4,2])
        def demazure_operator(self, element, reduced_word):
            r"""
            Return the application of Demazure operators `D_i` for `i` from
            ``reduced_word`` on ``element``.

            INPUT:

            - ``element`` -- an element of a free module indexed by the
              underlying crystal
            - ``reduced_word`` -- a reduced word of the Weyl group of the
              same type as the underlying crystal

            OUTPUT: an element of the free module indexed by the underlying crystal

            EXAMPLES::

                sage: T = crystals.Tableaux(['A',2], shape=[2,1])
                sage: C = CombinatorialFreeModule(QQ, T)                                # needs sage.modules
                sage: t = T.highest_weight_vector()
                sage: b = 2*C(t)                                                        # needs sage.modules
                sage: T.demazure_operator(b,[1,2,1])                                    # needs sage.modules
                2*B[[[1, 1], [2]]] + 2*B[[[1, 2], [2]]] + 2*B[[[1, 3], [2]]]
                + 2*B[[[1, 1], [3]]] + 2*B[[[1, 2], [3]]] + 2*B[[[1, 3], [3]]]
                + 2*B[[[2, 2], [3]]] + 2*B[[[2, 3], [3]]]

            The Demazure operator is idempotent::

                sage: # needs sage.modules
                sage: T = crystals.Tableaux("A1", shape=[4])
                sage: C = CombinatorialFreeModule(QQ, T)
                sage: b = C(T.module_generators[0]); b
                B[[[1, 1, 1, 1]]]
                sage: e = T.demazure_operator(b,[1]); e
                B[[[1, 1, 1, 1]]] + B[[[1, 1, 1, 2]]] + B[[[1, 1, 2, 2]]]
                + B[[[1, 2, 2, 2]]] + B[[[2, 2, 2, 2]]]
                sage: e == T.demazure_operator(e,[1])
                True
                sage: all(T.demazure_operator(T.demazure_operator(C(t),[1]),[1])
                ....:      == T.demazure_operator(C(t),[1]) for t in T)
                True
            """
            M = element.parent()
            for i in reversed(reduced_word):
                element = M.linear_combination((c.demazure_operator_simple(i), coeff)
                                               for c, coeff in element)
            return element

        def demazure_subcrystal(self, element, reduced_word, only_support=True):
            r"""
            Return the subcrystal corresponding to the application of
            Demazure operators `D_i` for `i` from ``reduced_word`` on
            ``element``.

            INPUT:

            - ``element`` -- an element of a free module indexed by the
              underlying crystal
            - ``reduced_word`` -- a reduced word of the Weyl group of the
              same type as the underlying crystal
            - ``only_support`` -- boolean (default: ``True``); only include
              arrows corresponding to the support of ``reduced_word``

            OUTPUT: the Demazure subcrystal

            EXAMPLES::

                sage: T = crystals.Tableaux(['A',2], shape=[2,1])
                sage: t = T.highest_weight_vector()
                sage: S = T.demazure_subcrystal(t, [1,2])
                sage: list(S)
                [[[1, 1], [2]], [[1, 2], [2]], [[1, 1], [3]],
                 [[1, 2], [3]], [[2, 2], [3]]]
                sage: S = T.demazure_subcrystal(t, [2,1])
                sage: list(S)
                [[[1, 1], [2]], [[1, 2], [2]], [[1, 1], [3]],
                 [[1, 3], [2]], [[1, 3], [3]]]

            We construct an example where we don't only want the arrows
            indicated by the support of the reduced word::

                sage: K = crystals.KirillovReshetikhin(['A',1,1], 1, 2)
                sage: mg = K.module_generator()
                sage: S = K.demazure_subcrystal(mg, [1])
                sage: S.digraph().edges(sort=True)
                [([[1, 1]], [[1, 2]], 1), ([[1, 2]], [[2, 2]], 1)]
                sage: S = K.demazure_subcrystal(mg, [1], only_support=False)
                sage: S.digraph().edges(sort=True)
                [([[1, 1]], [[1, 2]], 1),
                 ([[1, 2]], [[1, 1]], 0),
                 ([[1, 2]], [[2, 2]], 1),
                 ([[2, 2]], [[1, 2]], 0)]
            """
            from sage.combinat.free_module import CombinatorialFreeModule
            from sage.rings.rational_field import QQ
            C = CombinatorialFreeModule(QQ, self)
            D = self.demazure_operator(C(element), reduced_word)
            if only_support:
                index_set = tuple(frozenset(reduced_word))
            else:
                index_set = self.cartan_type().index_set()
            return self.subcrystal(contained=D.support(), generators=[element],
                                   index_set=index_set)

        def _test_stembridge_local_axioms(self, index_set=None, verbose=False, complete=False, **options):
            r"""
            This implements tests for the Stembridge local characterization
            on the finite crystal ``self``.

            The current implementation only uses the rules for simply-laced
            types.  Crystals of other types should still pass the test, but
            expansion of this test to non-simply laced type would be desirable.

            One can specify an index set smaller than the full index set of
            the crystal, using the option ``index_set``.

            Running with ``verbose=True`` will print each node for which a
            local axiom test applies.

            Running with ``complete=True`` will continue to run the test past
            the first failure of the local axioms.  This is probably only
            useful in conjunction with the verbose option, to see all places
            where the local axioms fail.

            EXAMPLES::

                sage: T = crystals.Tableaux(['A',3], shape=[2,1])
                sage: T._test_stembridge_local_axioms()
                True
                sage: T._test_stembridge_local_axioms(verbose=True)
                True
                sage: T._test_stembridge_local_axioms(index_set=[1,3])
                True

                sage: B = Crystals().example(choice='naive')
                sage: B._test_stembridge_local_axioms()
                Traceback (most recent call last):
                ...
                AssertionError: None
            """
            tester = self._tester(**options)
            goodness = True
            i = 0
            for x in self:
                goodness = x._test_stembridge_local_axioms(index_set, verbose)
                if not goodness and not complete:
                    tester.fail()
                i += 1
                if i > tester._max_runs:
                    return
            tester.assertTrue(goodness)
            return goodness

        def dual_equivalence_graph(self, X=None, index_set=None, directed=True):
            r"""
            Return the dual equivalence graph indexed by ``index_set``
            on the subset ``X`` of ``self``.

            Let `b \in B` be an element of weight `0`, so `\varepsilon_j(b)
            = \varphi_j(b)` for all `j \in I`, where `I` is the indexing
            set. We say `b'` is an `i`-elementary dual equivalence
            transformation of `b` (where `i \in I`) if

            * `\varepsilon_i(b) = 1` and `\varepsilon_{i-1}(b) = 0`, and
            * `b' = f_{i-1} f_i e_{i-1} e_i b`.

            We can do the inverse procedure by interchanging `i` and `i-1`
            above.

            .. NOTE::

                If the index set is not an ordered interval, we let
                `i - 1` mean the index appearing before `i` in `I`.

            This definition comes from [As2008]_ Section 4 (where our
            `\varphi_j(b)` and `\varepsilon_j(b)` are denoted by
            `\epsilon(b, j)` and `-\delta(b, j)`, respectively).

            The dual equivalence graph of `B` is defined to be the
            colored graph whose vertices are the elements of `B` of
            weight `0`, and whose edges of color `i` (for `i \in I`)
            connect pairs `\{ b, b' \}` such that `b'` is an
            `i`-elementary dual equivalence transformation of `b`.

            .. NOTE::

                This dual equivalence graph is a generalization of
                `\mathcal{G}\left(\mathcal{X}\right)` in [As2008]_
                Section 4 except we do not require
                `\varepsilon_i(b) = 0, 1` for all `i`.

            This definition can be generalized by choosing a subset `X`
            of the set of all vertices of `B` of weight `0`, and
            restricting the dual equivalence graph to the vertex set
            `X`.

            INPUT:

            - ``X`` -- (optional) the vertex set `X` (default:
              the whole set of vertices of ``self`` of weight `0`)
            - ``index_set`` -- (optional) the index set `I`
              (default: the whole index set of ``self``); this has
              to be a subset of the index set of ``self`` (as a list
              or tuple)
            - ``directed`` -- boolean (default: ``True``); whether to have the
              dual equivalence graph be directed, where the head of
              an edge `b - b'` is `b` and the tail is
              `b' = f_{i-1} f_i e_{i-1} e_i b`)

            .. SEEALSO::

                :meth:`sage.combinat.partition.Partition.dual_equivalence_graph`

            EXAMPLES::

                sage: T = crystals.Tableaux(['A',3], shape=[2,2])
                sage: G = T.dual_equivalence_graph()
                sage: G.edges(sort=True)
                [([[1, 3], [2, 4]], [[1, 2], [3, 4]], 2),
                 ([[1, 2], [3, 4]], [[1, 3], [2, 4]], 3)]
                sage: T = crystals.Tableaux(['A',4], shape=[3,2])
                sage: G = T.dual_equivalence_graph()
                sage: G.edges(sort=True)
                [([[1, 3, 5], [2, 4]], [[1, 3, 4], [2, 5]], 4),
                 ([[1, 3, 5], [2, 4]], [[1, 2, 5], [3, 4]], 2),
                 ([[1, 3, 4], [2, 5]], [[1, 2, 4], [3, 5]], 2),
                 ([[1, 2, 5], [3, 4]], [[1, 3, 5], [2, 4]], 3),
                 ([[1, 2, 4], [3, 5]], [[1, 2, 3], [4, 5]], 3),
                 ([[1, 2, 3], [4, 5]], [[1, 2, 4], [3, 5]], 4)]

                sage: T = crystals.Tableaux(['A',4], shape=[3,1])
                sage: G = T.dual_equivalence_graph(index_set=[1,2,3])
                sage: G.vertices(sort=True)
                [[[1, 3, 4], [2]], [[1, 2, 4], [3]], [[1, 2, 3], [4]]]
                sage: G.edges(sort=True)
                [([[1, 3, 4], [2]], [[1, 2, 4], [3]], 2),
                 ([[1, 2, 4], [3]], [[1, 2, 3], [4]], 3)]

            TESTS::

                sage: T = crystals.Tableaux(['A',4], shape=[3,1])
                sage: G = T.dual_equivalence_graph(index_set=[2,3])
                sage: G.edges(sort=True)
                [([[1, 2, 4], [3]], [[1, 2, 3], [4]], 3),
                 ([[2, 4, 5], [3]], [[2, 3, 5], [4]], 3)]
                sage: G.vertices(sort=True)
                [[[1, 3, 4], [2]],
                 [[1, 2, 4], [3]],
                 [[2, 4, 5], [3]],
                 [[1, 2, 3], [4]],
                 [[2, 3, 5], [4]],
                 [[1, 1, 1], [5]],
                 [[1, 1, 5], [5]],
                 [[1, 5, 5], [5]],
                 [[2, 3, 4], [5]]]
            """
            if index_set is None:
                index_set = self.index_set()

            def wt_zero(x):
                for i in index_set:
                    if x.epsilon(i) != x.phi(i):
                        return False
                return True

            if X is None:
                X = [x for x in self if wt_zero(x)]
                checker = lambda x: True
            elif any(not wt_zero(x) for x in X):
                raise ValueError("the elements are not all weight 0")
            else:
                checker = lambda x: x in X

            edges = []
            for x in X:
                for k, i in enumerate(index_set[1:]):
                    im = index_set[k]
                    if x.epsilon(i) == 1 and x.epsilon(im) == 0:
                        y = x.e(i).e(im).f(i).f(im)
                        if checker(y):
                            edges.append([x, y, i])
            from sage.graphs.digraph import DiGraph
            G = DiGraph([X, edges], format='vertices_and_edges', immutable=True)
            from sage.graphs.dot2tex_utils import have_dot2tex
            if have_dot2tex():
                G.set_latex_options(format='dot2tex', edge_labels=True,
                                    color_by_label=self.cartan_type()._index_set_coloring)
            return G

    class ElementMethods:

        def epsilon(self, i):
            r"""
            Return `\varepsilon_i` of ``self``.

            EXAMPLES::

                sage: C = crystals.Letters(['A',5])
                sage: C(1).epsilon(1)
                0
                sage: C(2).epsilon(1)
                1
            """
            assert i in self.index_set()
            x = self.e(i)
            eps = 0
            while x is not None:
                x = x.e(i)
                eps = eps + 1
            return eps

        def phi(self, i):
            r"""
            Return `\varphi_i` of ``self``.

            EXAMPLES::

                sage: C = crystals.Letters(['A',5])
                sage: C(1).phi(1)
                1
                sage: C(2).phi(1)
                0
            """
            assert i in self.index_set()
            x = self.f(i)
            phi = 0
            while x is not None:
                x = x.f(i)
                phi += 1
            return phi

        def weight(self):
            """
            Return the weight of this crystal element.

            EXAMPLES::

                sage: C = crystals.Letters(['A',5])
                sage: C(1).weight()
                (1, 0, 0, 0, 0, 0)
            """
            return self.Phi() - self.Epsilon()

        def demazure_operator_simple(self, i, ring=None):
            r"""
            Return the Demazure operator `D_i` applied to ``self``.

            INPUT:

            - ``i`` -- an element of the index set of the underlying crystal
            - ``ring`` -- (default: ``QQ``) a ring

            OUTPUT:

            An element of the ``ring``-free module indexed by the underlying
            crystal.

            Let `r = \langle \mathrm{wt}(b), \alpha^{\vee}_i \rangle`, then
            `D_i(b)` is defined as follows:

            - If `r \geq 0`, this returns the sum of the elements obtained
              from ``self`` by application of `f_i^k` for `0 \leq k \leq r`.
            - If `r < 0`, this returns the opposite of the sum of the
              elements obtained by application of `e_i^k` for `0 < k < -r`.

            REFERENCES:

            - [Li1995]_

            - [Ka1993]_

            EXAMPLES::

                sage: T = crystals.Tableaux(['A',2], shape=[2,1])
                sage: t = T(rows=[[1,2],[2]])
                sage: t.demazure_operator_simple(2)
                B[[[1, 2], [2]]] + B[[[1, 3], [2]]] + B[[[1, 3], [3]]]
                sage: t.demazure_operator_simple(2).parent()
                Algebra of The crystal of tableaux of type ['A', 2] and shape(s) [[2, 1]]
                        over Integer Ring

                sage: t.demazure_operator_simple(1)
                0

                sage: K = crystals.KirillovReshetikhin(['A',2,1],2,1)
                sage: t = K(rows=[[3],[2]])
                sage: t.demazure_operator_simple(0)
                B[[[1, 2]]] + B[[[2, 3]]]

            TESTS::

                sage: K = crystals.KirillovReshetikhin(['A',2,1],1,1)
                sage: x = K.an_element(); x
                [[1]]
                sage: x.demazure_operator_simple(0)
                0
                sage: x.demazure_operator_simple(0, ring = QQ).parent()
                Algebra of Kirillov-Reshetikhin crystal of type ['A', 2, 1] with (r,s)=(1,1)
                        over Rational Field
            """
            from sage.rings.integer_ring import ZZ
            if ring is None:
                ring = ZZ
            C = self.parent().algebra(ring)
            r = self.phi(i) - self.epsilon(i)
            if r >= 0:
                l = [self]
                element = self
                for k in range(r):
                    element = element.f(i)
                    l.append(element)
                return C.sum_of_monomials(l)
            else:
                l = []
                element = self
                for k in range(-r-1):
                    element = element.e(i)
                    l.append(element)
                return - C.sum_of_monomials(l)

        def stembridgeDelta_depth(self,i,j):
            r"""
            Return the difference in the `j`-depth of ``self`` and `e_i`
            of ``self``, where `i` and `j` are in the index set of the
            underlying crystal. This function is useful for checking the
            Stembridge local axioms for crystal bases.

            The `i`-depth of a crystal node `x` is `-\varepsilon_i(x)`.

            EXAMPLES::

                sage: T = crystals.Tableaux(['A',2], shape=[2,1])
                sage: t = T(rows=[[1,2],[2]])
                sage: t.stembridgeDelta_depth(1,2)
                0
                sage: s = T(rows=[[2,3],[3]])
                sage: s.stembridgeDelta_depth(1,2)
                -1
            """
            if self.e(i) is None:
                return 0
            return -self.e(i).epsilon(j) + self.epsilon(j)

        def stembridgeDelta_rise(self,i,j):
            r"""
            Return the difference in the `j`-rise of ``self`` and `e_i` of
            ``self``, where `i` and `j` are in the index set of the
            underlying crystal. This function is useful for checking the
            Stembridge local axioms for crystal bases.

            The `i`-rise of a crystal node `x` is `\varphi_i(x)`.

            EXAMPLES::

                sage: T = crystals.Tableaux(['A',2], shape=[2,1])
                sage: t = T(rows=[[1,2],[2]])
                sage: t.stembridgeDelta_rise(1,2)
                -1
                sage: s = T(rows=[[2,3],[3]])
                sage: s.stembridgeDelta_rise(1,2)
                0
            """
            if self.e(i) is None:
                return 0
            return self.e(i).phi(j) - self.phi(j)

        def stembridgeDel_depth(self,i,j):
            r"""
            Return the difference in the `j`-depth of ``self`` and `f_i` of
            ``self``, where `i` and `j` are in the index set of the
            underlying crystal. This function is useful for checking the
            Stembridge local axioms for crystal bases.

            The `i`-depth of a crystal node `x` is `\varepsilon_i(x)`.

            EXAMPLES::

                sage: T = crystals.Tableaux(['A',2], shape=[2,1])
                sage: t = T(rows=[[1,1],[2]])
                sage: t.stembridgeDel_depth(1,2)
                0
                sage: s = T(rows=[[1,3],[3]])
                sage: s.stembridgeDel_depth(1,2)
                -1
            """
            if self.f(i) is None:
                return 0
            return -self.epsilon(j) + self.f(i).epsilon(j)

        def stembridgeDel_rise(self,i,j):
            r"""
            Return the difference in the `j`-rise of ``self`` and `f_i` of
            ``self``, where `i` and `j` are in the index set of the
            underlying crystal. This function is useful for checking the
            Stembridge local axioms for crystal bases.

            The `i`-rise of a crystal node `x` is `\varphi_i(x)`.

            EXAMPLES::

                sage: T = crystals.Tableaux(['A',2], shape=[2,1])
                sage: t = T(rows=[[1,1],[2]])
                sage: t.stembridgeDel_rise(1,2)
                -1
                sage: s = T(rows=[[1,3],[3]])
                sage: s.stembridgeDel_rise(1,2)
                0
            """
            if self.f(i) is None:
                return 0
            return self.phi(j)-self.f(i).phi(j)

        def stembridgeTriple(self,i,j):
            r"""
            Let `A` be the Cartan matrix of the crystal, `x` a crystal element,
            and let `i` and `j` be in the index set of the crystal.
            Further, set
            ``b=stembridgeDelta_depth(x,i,j)``, and
            ``c=stembridgeDelta_rise(x,i,j))``.
            If ``x.e(i)`` is non-empty, this function returns the triple
            `( A_{ij}, b, c )`; otherwise it returns ``None``.
            By the Stembridge local characterization of crystal bases,
            one should have `A_{ij}=b+c`.

            EXAMPLES::

                sage: T = crystals.Tableaux(['A',2], shape=[2,1])
                sage: t = T(rows=[[1,1],[2]])
                sage: t.stembridgeTriple(1,2)
                sage: s = T(rows=[[1,2],[2]])
                sage: s.stembridgeTriple(1,2)
                (-1, 0, -1)

                sage: T = crystals.Tableaux(['B',2], shape=[2,1])
                sage: t = T(rows=[[1,2],[2]])
                sage: t.stembridgeTriple(1,2)
                (-2, 0, -2)
                sage: s = T(rows=[[-1,-1],[0]])
                sage: s.stembridgeTriple(1,2)
                (-2, -2, 0)
                sage: u = T(rows=[[0,2],[1]])
                sage: u.stembridgeTriple(1,2)
                (-2, -1, -1)
            """
            if self.e(i) is None:
                return None
            b = self.stembridgeDelta_depth(i,j)
            c = self.stembridgeDelta_rise(i,j)
            dd = self.cartan_type().dynkin_diagram()
            a = dd[j,i]
            return (a, b, c)

        def _test_stembridge_local_axioms(self, index_set=None, verbose=False, **options):
            r"""
            This implements tests for the Stembridge local characterization
            on the element of a crystal ``self``.

            The current implementation only uses the axioms for simply-laced
            types.  Crystals of other types should still pass the test, but
            in non-simply-laced types, passing is not a guarantee that the
            crystal arises from a representation.

            One can specify an index set smaller than the full index set of
            the crystal, using the option ``index_set``.

            Running with ``verbose=True`` will print warnings when a test fails.

            REFERENCES:

            - [Ste2003]_

            EXAMPLES::

                sage: T = crystals.Tableaux(['A',2], shape=[2,1])
                sage: t = T(rows=[[1,1],[2]])
                sage: t._test_stembridge_local_axioms()
                True
                sage: t._test_stembridge_local_axioms(index_set=[1,3])
                True
                sage: t._test_stembridge_local_axioms(verbose=True)
                True
            """
            tester = self._tester(**options)
            goodness = True
            if index_set is None:
                index_set = self.index_set()

            from sage.combinat.subset import Subsets

            for (i,j) in Subsets(index_set, 2):
                if self.e(i) is not None and self.e(j) is not None:
                    triple = self.stembridgeTriple(i,j)
                    #Test axioms P3 and P4.
                    if not triple[0] == triple[1]+triple[2] or triple[1] > 0 or triple[2] > 0:
                        if verbose:
                            print('Warning: Failed axiom P3 or P4 at vector ', self, 'i,j=', i, j, 'Stembridge triple:', self.stembridgeTriple(i, j))
                            goodness = False
                        else:
                            tester.fail()
                    if self.stembridgeDelta_depth(i,j) == 0:
                        #check E_i E_j(x)= E_j E_i(x)
                        if self.e(i).e(j) != self.e(j).e(i) or self.e(i).e(j).stembridgeDel_rise(j, i) != 0:
                            if verbose:
                                print('Warning: Failed axiom P5 at: vector ', self, 'i,j=', i, j, 'Stembridge triple:', self.stembridgeTriple(i, j))
                                goodness = False
                            else:
                                tester.fail()
                    if self.stembridgeDelta_depth(i,j) == -1 and self.stembridgeDelta_depth(j,i) == -1:
                        #check E_i E_j^2 E_i (x)= E_j E_i^2 E_j (x)
                        y1 = self.e(j).e(i).e(i).e(j)
                        y2 = self.e(j).e(i).e(i).e(j)
                        a = y1.stembridgeDel_rise(j, i)
                        b = y2.stembridgeDel_rise(i, j)
                        if y1 != y2 or a != -1 or b != -1:
                            if verbose:
                                print('Warning: Failed axiom P6 at: vector ', self, 'i,j=', i, j, 'Stembridge triple:', self.stembridgeTriple(i, j))
                                goodness = False
                            else:
                                tester.fail()
            tester.assertTrue(goodness)
            return goodness

        def dual_equivalence_class(self, index_set=None):
            r"""
            Return the dual equivalence class indexed by ``index_set``
            of ``self``.

            The dual equivalence class of an element `b \in B`
            is the set of all elements of `B` reachable from
            `b` via sequences of `i`-elementary dual equivalence
            relations (i.e., `i`-elementary dual equivalence
            transformations and their inverses) for `i` in the index
            set of `B`.

            For this to be well-defined, the element `b` has to be
            of weight `0` with respect to `I`; that is, we need to have
            `\varepsilon_j(b) = \varphi_j(b)` for all `j \in I`.

            See [As2008]_. See also :meth:`dual_equivalence_graph` for
            a definition of `i`-elementary dual equivalence
            transformations.

            INPUT:

            - ``index_set`` -- (optional) the index set `I`
              (default: the whole index set of the crystal); this has
              to be a subset of the index set of the crystal (as a list
              or tuple)

            OUTPUT:

            The dual equivalence class of ``self`` indexed by the
            subset ``index_set``. This class is returned as an
            undirected edge-colored multigraph. The color of an edge
            is the index `i` of the dual equivalence relation it
            encodes.

            .. SEEALSO::

                - :meth:`~sage.categories.regular_crystals.RegularCrystals.ParentMethods.dual_equivalence_graph`
                - :meth:`sage.combinat.partition.Partition.dual_equivalence_graph`

            EXAMPLES::

                sage: T = crystals.Tableaux(['A',3], shape=[2,2])
                sage: G = T(2,1,4,3).dual_equivalence_class()
                sage: G.edges(sort=True)
                [([[1, 3], [2, 4]], [[1, 2], [3, 4]], 2),
                 ([[1, 3], [2, 4]], [[1, 2], [3, 4]], 3)]
                sage: T = crystals.Tableaux(['A',4], shape=[3,2])
                sage: G = T(2,1,4,3,5).dual_equivalence_class()
                sage: G.edges(sort=True)
                [([[1, 3, 5], [2, 4]], [[1, 3, 4], [2, 5]], 4),
                 ([[1, 3, 5], [2, 4]], [[1, 2, 5], [3, 4]], 2),
                 ([[1, 3, 5], [2, 4]], [[1, 2, 5], [3, 4]], 3),
                 ([[1, 3, 4], [2, 5]], [[1, 2, 4], [3, 5]], 2),
                 ([[1, 2, 4], [3, 5]], [[1, 2, 3], [4, 5]], 3),
                 ([[1, 2, 4], [3, 5]], [[1, 2, 3], [4, 5]], 4)]
            """
            if index_set is None:
                index_set = self.index_set()

            for i in index_set:
                if self.epsilon(i) != self.phi(i):
                    raise ValueError("the element is not weight 0")

            visited = set()
            todo = {self}
            edges = []
            while todo:
                x = todo.pop()
                visited.add(x)
                for k, i in enumerate(index_set[1:]):
                    im = index_set[k]
                    if x.epsilon(i) == 1 and x.epsilon(im) == 0:
                        y = x.e(i).e(im).f(i).f(im)
                        if [y, x, i] not in edges:
                            edges.append([x, y, i])
                        if y not in visited:
                            todo.add(y)
                    if x.epsilon(i) == 0 and x.epsilon(im) == 1:
                        y = x.e(im).e(i).f(im).f(i)
                        if [y, x, i] not in edges:
                            edges.append([x, y, i])
                        if y not in visited:
                            todo.add(y)
            from sage.graphs.graph import Graph
            G = Graph([visited, edges], format='vertices_and_edges',
                      immutable=True, multiedges=True)
            from sage.graphs.dot2tex_utils import have_dot2tex
            if have_dot2tex():
                G.set_latex_options(format='dot2tex', edge_labels=True,
                                    color_by_label=self.cartan_type()._index_set_coloring)
            return G

    class TensorProducts(TensorProductsCategory):
        """
        The category of regular crystals constructed by tensor
        product of regular crystals.
        """
        @cached_method
        def extra_super_categories(self):
            """
            EXAMPLES::

                sage: RegularCrystals().TensorProducts().extra_super_categories()
                [Category of regular crystals]
            """
            return [self.base_category()]



# sage_setup: distribution = sagemath-categories
"""
Finitely Generated Lambda bracket Algebras

AUTHORS:

- Reimundo Heluani (2020-08-21): Initial implementation.
"""

#******************************************************************************
#       Copyright (C) 2020 Reimundo Heluani <heluani@potuz.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from sage.categories.category_with_axiom import CategoryWithAxiom_over_base_ring
from sage.categories.graded_modules import GradedModulesCategory
from sage.categories.lambda_bracket_algebras import LambdaBracketAlgebras


class FinitelyGeneratedLambdaBracketAlgebras(CategoryWithAxiom_over_base_ring):
    """
    The category of finitely generated lambda bracket algebras.

    EXAMPLES::

        sage: from sage.categories.lambda_bracket_algebras import LambdaBracketAlgebras
        sage: LambdaBracketAlgebras(QQbar).FinitelyGenerated()                          # needs sage.rings.number_field
        Category of finitely generated lambda bracket algebras over Algebraic Field
    """
    _base_category_class_and_axiom = (LambdaBracketAlgebras, "FinitelyGeneratedAsLambdaBracketAlgebra")

    class ParentMethods:
        def ngens(self):
            r"""
            The number of generators of this Lie conformal algebra.

            EXAMPLES::

                sage: Vir = lie_conformal_algebras.Virasoro(QQ)                         # needs sage.combinat sage.modules
                sage: Vir.ngens()                                                       # needs sage.combinat sage.modules
                2

                sage: V = lie_conformal_algebras.Affine(QQ, 'A2')                       # needs sage.combinat sage.modules
                sage: V.ngens()                                                         # needs sage.combinat sage.modules
                9
            """
            return len(self.gens())

        def gen(self,i):
            r"""
            The ``i``-th generator of this Lie conformal algebra.

            EXAMPLES::

                sage: # needs sage.combinat sage.modules
                sage: V = lie_conformal_algebras.Affine(QQ, 'A1')
                sage: V.gens()
                (B[alpha[1]], B[alphacheck[1]], B[-alpha[1]], B['K'])
                sage: V.gen(0)
                B[alpha[1]]
                sage: V.1
                B[alphacheck[1]]
            """
            return self.gens()[i]

        def some_elements(self):
            """
            Some elements of this Lie conformal algebra.

            This method returns a list with elements containing at
            least the generators.

            EXAMPLES::

                sage: V = lie_conformal_algebras.Affine(QQ, 'A1',                       # needs sage.combinat sage.modules
                ....:                                   names=('e', 'h', 'f'))
                sage: V.some_elements()                                                 # needs sage.combinat sage.modules
                [e, h, f, K, ...]
                sage: all(v.parent() is V for v in V.some_elements())                   # needs sage.combinat sage.modules
                True
            """
            S = list(self.gens())
            from sage.misc.misc import some_tuples
            for x,y in some_tuples(S, 2, 0, max_samples=self.ngens()):
                S.append(x.T() + 2*y.T(2))
            return S

    class Graded(GradedModulesCategory):
        """
        The category of H-graded finitely generated Lie conformal algebras.

        EXAMPLES::

            sage: LieConformalAlgebras(QQbar).FinitelyGenerated().Graded()              # needs sage.rings.number_field
            Category of H-graded finitely generated Lie conformal algebras
             over Algebraic Field
        """
        def _repr_object_names(self):
            """
            The names of the objects of ``self``.

            EXAMPLES::

                sage: LieConformalAlgebras(QQbar).WithBasis().FinitelyGenerated().Graded()          # needs sage.rings.number_field
                Category of H-graded finitely generated Lie conformal algebras with basis over Algebraic Field
            """
            return "H-graded {}".format(self.base_category()._repr_object_names())



# sage_setup: distribution = sagemath-categories
r"""
Principal ideal domains
"""
# ****************************************************************************
#  Copyright (C) 2008 Teresa Gomez-Diaz (CNRS) <Teresa.Gomez-Diaz@univ-mlv.fr>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  https://www.gnu.org/licenses/
# *****************************************************************************

from sage.categories.category_singleton import Category_singleton
from sage.categories.unique_factorization_domains import UniqueFactorizationDomains


class PrincipalIdealDomains(Category_singleton):
    """
    The category of (constructive) principal ideal domains.

    By constructive, we mean that a single generator can be
    constructively found for any ideal given by a finite set of
    generators. Note that this constructive definition only implies
    that finitely generated ideals are principal. It is not clear what
    we would mean by an infinitely generated ideal.

    EXAMPLES::

      sage: PrincipalIdealDomains()
      Category of principal ideal domains
      sage: PrincipalIdealDomains().super_categories()
      [Category of unique factorization domains]

    See also :wikipedia:`Principal_ideal_domain`

    TESTS::

        sage: TestSuite(PrincipalIdealDomains()).run()
    """

    def super_categories(self):
        """
        EXAMPLES::

            sage: PrincipalIdealDomains().super_categories()
            [Category of unique factorization domains]
        """
        return [UniqueFactorizationDomains()]

    def additional_structure(self):
        """
        Return ``None``.

        Indeed, the category of principal ideal domains defines no
        additional structure: a ring morphism between two principal
        ideal domains is a principal ideal domain morphism.

        EXAMPLES::

            sage: PrincipalIdealDomains().additional_structure()
        """
        return None

    class ParentMethods:
        def _test_gcd_vs_xgcd(self, **options):
            r"""
            Check that gcd and xgcd are compatible if implemented.

            This test will prevent things like :issue:`17671` to happen again.

            TESTS::

                sage: ZZ._test_gcd_vs_xgcd()
                sage: QQ._test_gcd_vs_xgcd()
                sage: QQ['x']._test_gcd_vs_xgcd()
                sage: QQbar['x']._test_gcd_vs_xgcd()                                    # needs sage.rings.number_field
                sage: RR._test_gcd_vs_xgcd()
                sage: RR['x']._test_gcd_vs_xgcd()

            A slightly more involved example of polynomial ring with a non UFD
            base ring::

                sage: # needs sage.rings.number_field
                sage: K = QuadraticField(5)
                sage: O = K.maximal_order()
                sage: O in UniqueFactorizationDomains()
                False
                sage: R = PolynomialRing(O, 'x')
                sage: F = R.fraction_field()
                sage: F in PrincipalIdealDomains()
                True
                sage: F._test_gcd_vs_xgcd()
            """
            tester = self._tester(**options)
            elts = list(tester.some_elements())

            # there are some strange things in Sage doctests... so it is better
            # to cut the list in order to avoid lists of size 531441.
            elts = elts[:10]
            pairs = [(x, y) for x in elts for y in elts]

            try:
                xgcds = [x.xgcd(y) for x, y in pairs]
            except (AttributeError, NotImplementedError):
                return

            has_gcd = True
            try:
                gcds = [x.gcd(y) for x, y in pairs]
            except (AttributeError, NotImplementedError):
                has_gcd = False

            tester.assertTrue(has_gcd,
                    "The ring {} provides a xgcd but no gcd".format(self))
            for (x, y), gcd, xgcd in zip(pairs, gcds, xgcds):
                tester.assertTrue(gcd.parent() == self,
                        "The parent of the gcd is {} for element of {}".format(
                            gcd.parent(), self))
                tester.assertTrue(xgcd[0].parent() == self and
                                  xgcd[1].parent() == self == xgcd[2].parent(),
                                  "The parent of output in xgcd is different from "
                                  "the parent of input for elements in {}".format(self))
                tester.assertTrue(gcd == xgcd[0],
                        "The methods gcd and xgcd disagree on {}:\n"
                        "  gcd({},{}) = {}\n"
                        " xgcd({},{}) = {}\n".format(self, x, y, gcd, x, y, xgcd))

        def is_noetherian(self) -> bool:
            """
            Every principal ideal domain is Noetherian, so we return ``True``.

            EXAMPLES::

                sage: Zp(5).is_noetherian()                                                 # needs sage.rings.padics
                True
            """
            return True

        def class_group(self):
            """
            Return the trivial group, since the class group of a PID is trivial.

            EXAMPLES::

                sage: QQ.class_group()                                                      # needs sage.groups
                Trivial Abelian group
            """
            from sage.groups.abelian_gps.abelian_group import AbelianGroup
            return AbelianGroup([])

        def gcd(self, x, y, coerce=True):
            r"""
            Return the greatest common divisor of ``x`` and ``y``, as elements
            of ``self``.

            EXAMPLES:

            The integers are a principal ideal domain and hence a GCD domain::

                sage: ZZ.gcd(42, 48)
                6
                sage: 42.factor(); 48.factor()
                2 * 3 * 7
                2^4 * 3
                sage: ZZ.gcd(2^4*7^2*11, 2^3*11*13)
                88
                sage: 88.factor()
                2^3 * 11

            In a field, any nonzero element is a GCD of any nonempty set
            of nonzero elements. In previous versions, Sage used to return
            1 in the case of the rational field. However, since :issue:`10771`,
            the rational field is considered as the
            *fraction field* of the integer ring. For the fraction field
            of an integral domain that provides both GCD and LCM, it is
            possible to pick a GCD that is compatible with the GCD of the
            base ring::

                sage: QQ.gcd(ZZ(42), ZZ(48)); type(QQ.gcd(ZZ(42), ZZ(48)))
                6
                <class 'sage.rings.rational.Rational'>
                sage: QQ.gcd(1/2, 1/3)
                1/6

            Polynomial rings over fields are GCD domains as well. Here is a simple
            example over the ring of polynomials over the rationals as well as
            over an extension ring. Note that ``gcd`` requires x and y to be
            coercible::

                sage: # needs sage.rings.number_field
                sage: R.<x> = PolynomialRing(QQ)
                sage: S.<a> = NumberField(x^2 - 2, 'a')
                sage: f = (x - a)*(x + a); g = (x - a)*(x^2 - 2)
                sage: print(f); print(g)
                x^2 - 2
                x^3 - a*x^2 - 2*x + 2*a
                sage: f in R
                True
                sage: g in R
                False
                sage: R.gcd(f, g)
                Traceback (most recent call last):
                ...
                TypeError: Unable to coerce 2*a to a rational
                sage: R.base_extend(S).gcd(f,g)
                x^2 - 2
                sage: R.base_extend(S).gcd(f, (x - a)*(x^2 - 3))
                x - a
            """
            if coerce:
                x = self(x)
                y = self(y)
            return x.gcd(y)

        def content(self, x, y, coerce=True):
            r"""
            Return the content of `x` and `y`.

            This is the unique element `c` of
            ``self`` such that `x/c` and `y/c`
            are coprime and integral.

            EXAMPLES::

                sage: QQ.content(ZZ(42), ZZ(48)); type(QQ.content(ZZ(42), ZZ(48)))
                6
                <class 'sage.rings.rational.Rational'>
                sage: QQ.content(1/2, 1/3)
                1/6
                sage: factor(1/2); factor(1/3); factor(1/6)
                2^-1
                3^-1
                2^-1 * 3^-1
                sage: a = (2*3)/(7*11); b = (13*17)/(19*23)
                sage: factor(a); factor(b); factor(QQ.content(a,b))
                2 * 3 * 7^-1 * 11^-1
                13 * 17 * 19^-1 * 23^-1
                7^-1 * 11^-1 * 19^-1 * 23^-1

            Note the changes to the second entry::

                sage: c = (2*3)/(7*11); d = (13*17)/(7*19*23)
                sage: factor(c); factor(d); factor(QQ.content(c,d))
                2 * 3 * 7^-1 * 11^-1
                7^-1 * 13 * 17 * 19^-1 * 23^-1
                7^-1 * 11^-1 * 19^-1 * 23^-1
                sage: e = (2*3)/(7*11); f = (13*17)/(7^3*19*23)
                sage: factor(e); factor(f); factor(QQ.content(e,f))
                2 * 3 * 7^-1 * 11^-1
                7^-3 * 13 * 17 * 19^-1 * 23^-1
                7^-3 * 11^-1 * 19^-1 * 23^-1
            """
            if coerce:
                x = self(x)
                y = self(y)
            return x.content(y)

        def _ideal_class_(self, n=0):
            """
            Ideals in PIDs have their own special class.

            EXAMPLES::

                sage: ZZ._ideal_class_()
                <class 'sage.rings.ideal.Ideal_pid'>
            """
            from sage.rings.ideal import Ideal_pid
            return Ideal_pid

    class ElementMethods:
        pass



# sage_setup: distribution = sagemath-categories
r"""
Filtered Modules

A *filtered module* over a ring `R` with a totally ordered
indexing set `I` (typically `I = \NN`) is an `R`-module `M` equipped
with a family `(F_i)_{i \in I}` of `R`-submodules satisfying
`F_i \subseteq F_j` for all `i,j \in I` having `i \leq j`, and
`M = \bigcup_{i \in I} F_i`. This family is called a *filtration*
of the given module `M`.

.. TODO::

    Implement a notion for decreasing filtrations: where `F_j \subseteq F_i`
    when `i \leq j`.

.. TODO::

    Implement filtrations for all concrete categories.

.. TODO::

    Implement `\operatorname{gr}` as a functor.
"""
# ****************************************************************************
#  Copyright (C) 2014 Travis Scrimshaw <tscrim at ucdavis.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  https://www.gnu.org/licenses/
# *****************************************************************************

from sage.misc.cachefunc import cached_method
from sage.categories.category_types import Category_over_base_ring
from sage.categories.category_with_axiom import CategoryWithAxiom_over_base_ring
from sage.categories.covariant_functorial_construction import RegressiveCovariantConstructionCategory


class FilteredModulesCategory(RegressiveCovariantConstructionCategory, Category_over_base_ring):
    def __init__(self, base_category):
        """
        EXAMPLES::

            sage: C = Algebras(QQ).Filtered()
            sage: C
            Category of filtered algebras over Rational Field
            sage: C.base_category()
            Category of algebras over Rational Field
            sage: sorted(C.super_categories(), key=str)
            [Category of algebras over Rational Field,
             Category of filtered vector spaces over Rational Field]

            sage: AlgebrasWithBasis(QQ).Filtered().base_ring()
            Rational Field
            sage: HopfAlgebrasWithBasis(QQ).Filtered().base_ring()
            Rational Field
        """
        super().__init__(base_category, base_category.base_ring())

    _functor_category = "Filtered"

    def _repr_object_names(self):
        """
        EXAMPLES::

            sage: AlgebrasWithBasis(QQ).Filtered()  # indirect doctest
            Category of filtered algebras with basis over Rational Field
        """
        return "filtered {}".format(self.base_category()._repr_object_names())


class FilteredModules(FilteredModulesCategory):
    r"""
    The category of filtered modules over a given ring `R`.

    A *filtered module* over a ring `R` with a totally ordered
    indexing set `I` (typically `I = \NN`) is an `R`-module `M` equipped
    with a family `(F_i)_{i \in I}` of `R`-submodules satisfying
    `F_i \subseteq F_j` for all `i,j \in I` having `i \leq j`, and
    `M = \bigcup_{i \in I} F_i`. This family is called a *filtration*
    of the given module `M`.

    EXAMPLES::

        sage: Modules(ZZ).Filtered()
        Category of filtered modules over Integer Ring
        sage: Modules(ZZ).Filtered().super_categories()
        [Category of modules over Integer Ring]

    TESTS::

        sage: TestSuite(Modules(ZZ).Filtered()).run()

    REFERENCES:

    - :wikipedia:`Filtration_(mathematics)`
    """
    def extra_super_categories(self):
        r"""
        Add :class:`VectorSpaces` to the super categories of ``self`` if
        the base ring is a field.

        EXAMPLES::

            sage: Modules(QQ).Filtered().is_subcategory(VectorSpaces(QQ))
            True
            sage: Modules(ZZ).Filtered().extra_super_categories()
            []

        This makes sure that ``Modules(QQ).Filtered()`` returns an
        instance of :class:`FilteredModules` and not a join category of
        an instance of this class and of ``VectorSpaces(QQ)``::

            sage: type(Modules(QQ).Filtered())
            <class 'sage.categories.vector_spaces.VectorSpaces.Filtered_with_category'>

        .. TODO::

            Get rid of this workaround once there is a more systematic
            approach for the alias ``Modules(QQ)`` -> ``VectorSpaces(QQ)``.
            Probably the latter should be a category with axiom, and
            covariant constructions should play well with axioms.
        """
        from sage.categories.modules import Modules
        from sage.categories.fields import Fields
        base_ring = self.base_ring()
        if base_ring in Fields():
            return [Modules(base_ring)]
        else:
            return []

    class SubcategoryMethods:

        @cached_method
        def Connected(self):
            r"""
            Return the full subcategory of the connected objects of ``self``.

            A filtered `R`-module `M` with filtration
            `(F_0, F_1, F_2, \ldots)` (indexed by `\NN`)
            is said to be *connected* if `F_0` is isomorphic
            to `R`.

            EXAMPLES::

                sage: Modules(ZZ).Filtered().Connected()
                Category of filtered connected modules over Integer Ring
                sage: Coalgebras(QQ).Filtered().Connected()
                Category of filtered connected coalgebras over Rational Field
                sage: AlgebrasWithBasis(QQ).Filtered().Connected()
                Category of filtered connected algebras with basis over Rational Field

            TESTS::

                sage: TestSuite(Modules(ZZ).Filtered().Connected()).run()
                sage: Coalgebras(QQ).Filtered().Connected.__module__
                'sage.categories.filtered_modules'
            """
            return self._with_axiom("Connected")

    class Connected(CategoryWithAxiom_over_base_ring):
        pass



# sage_setup: distribution = sagemath-categories
r"""
Hopf algebras with basis
"""
# ****************************************************************************
#  Copyright (C) 2008 Teresa Gomez-Diaz (CNRS) <Teresa.Gomez-Diaz@univ-mlv.fr>
#  Copyright (C) 2008-2011 Nicolas M. Thiery <nthiery at users.sf.net>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  https://www.gnu.org/licenses/
# *****************************************************************************

from sage.categories.category_with_axiom import CategoryWithAxiom_over_base_ring
from sage.categories.tensor import TensorProductsCategory
from sage.misc.abstract_method import abstract_method
from sage.misc.cachefunc import cached_method
from sage.misc.lazy_attribute import lazy_attribute
from sage.misc.lazy_import import LazyImport


class HopfAlgebrasWithBasis(CategoryWithAxiom_over_base_ring):
    """
    The category of Hopf algebras with a distinguished basis.

    EXAMPLES::

        sage: C = HopfAlgebrasWithBasis(QQ)
        sage: C
        Category of Hopf algebras with basis over Rational Field
        sage: C.super_categories()
        [Category of Hopf algebras over Rational Field,
         Category of bialgebras with basis over Rational Field]

    We now show how to use a simple Hopf algebra, namely the group algebra of the dihedral group
    (see also AlgebrasWithBasis)::

        sage: A = C.example(); A                                                        # needs sage.groups
        An example of Hopf algebra with basis: the group algebra of the
         Dihedral group of order 6 as a permutation group over Rational Field
        sage: A.rename('A')                                                             # needs sage.groups
        sage: A.category()                                                              # needs sage.groups
        Category of finite dimensional Hopf algebras with basis over Rational Field

        sage: A.one_basis()                                                             # needs sage.groups
        ()
        sage: A.one()                                                                   # needs sage.groups
        B[()]

        sage: A.base_ring()                                                             # needs sage.groups
        Rational Field
        sage: A.basis().keys()                                                          # needs sage.groups
        Dihedral group of order 6 as a permutation group

        sage: # needs sage.groups
        sage: [a,b] = A.algebra_generators()
        sage: a, b
        (B[(1,2,3)], B[(1,3)])
        sage: a^3, b^2
        (B[()], B[()])
        sage: a*b
        B[(1,2)]

        sage: A.product           # todo: not quite ...                                 # needs sage.groups
        <bound method MagmaticAlgebras.WithBasis.ParentMethods._product_from_product_on_basis_multiply of A>
        sage: A.product(b, b)                                                           # needs sage.groups
        B[()]

        sage: A.zero().coproduct()                                                      # needs sage.groups
        0
        sage: A.zero().coproduct().parent()                                             # needs sage.groups
        A # A
        sage: a.coproduct()                                                             # needs sage.groups
        B[(1,2,3)] # B[(1,2,3)]

        sage: TestSuite(A).run(verbose=True)                                            # needs sage.groups
        running ._test_additive_associativity() . . . pass
        running ._test_an_element() . . . pass
        running ._test_antipode() . . . pass
        running ._test_associativity() . . . pass
        running ._test_cardinality() . . . pass
        running ._test_category() . . . pass
        running ._test_characteristic() . . . pass
        running ._test_construction() . . . pass
        running ._test_distributivity() . . . pass
        running ._test_elements() . . .
          Running the test suite of self.an_element()
          running ._test_category() . . . pass
          running ._test_eq() . . . pass
          running ._test_monomial_coefficients() . . . pass
          running ._test_new() . . . pass
          running ._test_nonzero_equal() . . . pass
          running ._test_not_implemented_methods() . . . pass
          running ._test_pickling() . . . pass
          pass
        running ._test_elements_eq_reflexive() . . . pass
        running ._test_elements_eq_symmetric() . . . pass
        running ._test_elements_eq_transitive() . . . pass
        running ._test_elements_neq() . . . pass
        running ._test_eq() . . . pass
        running ._test_new() . . . pass
        running ._test_not_implemented_methods() . . . pass
        running ._test_one() . . . pass
        running ._test_pickling() . . . pass
        running ._test_prod() . . . pass
        running ._test_some_elements() . . . pass
        running ._test_zero() . . . pass
        sage: A.__class__                                                               # needs sage.groups
        <class 'sage.categories.examples.hopf_algebras_with_basis.MyGroupAlgebra_with_category'>
        sage: A.element_class                                                           # needs sage.groups
        <class 'sage.categories.examples.hopf_algebras_with_basis.MyGroupAlgebra_with_category.element_class'>

    Let us look at the code for implementing A::

        sage: A??                               # not implemented                       # needs sage.groups

    TESTS::

        sage: TestSuite(A).run()                                                        # needs sage.groups
        sage: TestSuite(C).run()
    """

    def example(self, G=None):
        """
        Return an example of algebra with basis::

            sage: HopfAlgebrasWithBasis(QQ['x']).example()                              # needs sage.groups
            An example of Hopf algebra with basis: the group algebra of the
            Dihedral group of order 6 as a permutation group
            over Univariate Polynomial Ring in x over Rational Field

        An other group can be specified as optional argument::

            sage: HopfAlgebrasWithBasis(QQ).example(SymmetricGroup(4))                  # needs sage.groups
            An example of Hopf algebra with basis: the group algebra of the
            Symmetric group of order 4! as a permutation group over Rational Field
        """
        from sage.categories.examples.hopf_algebras_with_basis import MyGroupAlgebra
        from sage.groups.perm_gps.permgroup_named import DihedralGroup
        if G is None:
            G = DihedralGroup(3)
        return MyGroupAlgebra(self.base_ring(), G)

#     This is only correct in the finite dimensional / graded case
#     def dual(self):
#         """
#         Returns the dual category

#         EXAMPLES:

#         The category of Hopf algebras over any field is self dual::

#             sage: C = HopfAlgebrasWithBasis(QQ)
#             sage: C.dual()
#             Category of Hopf algebras with basis over Rational Field
#         """
#         return self

    FiniteDimensional = LazyImport('sage.categories.finite_dimensional_hopf_algebras_with_basis',
                                   'FiniteDimensionalHopfAlgebrasWithBasis')
    Filtered = LazyImport('sage.categories.filtered_hopf_algebras_with_basis',
                          'FilteredHopfAlgebrasWithBasis')
    Graded = LazyImport('sage.categories.graded_hopf_algebras_with_basis',
                        'GradedHopfAlgebrasWithBasis')
    Super = LazyImport('sage.categories.super_hopf_algebras_with_basis',
                       'SuperHopfAlgebrasWithBasis')

    class ParentMethods:

        @abstract_method(optional=True)
        def antipode_on_basis(self, x):
            """
            The antipode of the Hopf algebra on the basis (optional).

            INPUT:

            - ``x`` -- an index of an element of the basis of ``self``

            Returns the antipode of the basis element indexed by ``x``.

            If this method is implemented, then :meth:`antipode` is defined
            from this by linearity.

            EXAMPLES::

                sage: # needs sage.groups
                sage: A = HopfAlgebrasWithBasis(QQ).example()
                sage: W = A.basis().keys(); W
                Dihedral group of order 6 as a permutation group
                sage: w = W.gen(0); w
                (1,2,3)
                sage: A.antipode_on_basis(w)
                B[(1,3,2)]
            """

        @lazy_attribute
        def antipode(self):
            """
            The antipode of this Hopf algebra.

            If :meth:`.antipode_basis` is available, this constructs the
            antipode morphism from ``self`` to ``self`` by extending it by
            linearity. Otherwise, :meth:`self.antipode_by_coercion` is used, if
            available.

            EXAMPLES::

                sage: # needs sage.groups
                sage: A = HopfAlgebrasWithBasis(ZZ).example(); A
                An example of Hopf algebra with basis: the group algebra of the
                 Dihedral group of order 6 as a permutation group over Integer Ring
                sage: A = HopfAlgebrasWithBasis(QQ).example()
                sage: [a,b] = A.algebra_generators()
                sage: a, A.antipode(a)
                (B[(1,2,3)], B[(1,3,2)])
                sage: b, A.antipode(b)
                (B[(1,3)], B[(1,3)])

            TESTS::

                sage: all(A.antipode(x) * x == A.one() for x in A.basis())              # needs sage.groups
                True
            """
            if self.antipode_on_basis is not NotImplemented:
                # Should give the information that this is an anti-morphism of algebra
                return self._module_morphism(self.antipode_on_basis,
                                             codomain=self)
            elif hasattr(self, "antipode_by_coercion"):
                return self.antipode_by_coercion

        def _test_antipode(self, **options):
            r"""
            Test the antipode.

            An *antipode* `S` of a Hopf algebra is a linear endomorphism of the
            Hopf algebra that satisfies the following conditions (see
            :wikipedia:`HopfAlgebra`).

            - If `\mu` and `\Delta` denote the product and coproduct of the
              Hopf algebra, respectively, then `S` satisfies

              .. MATH::

                  \mu \circ (S \tensor 1) \circ \Delta = unit \circ counit
                  \mu \circ (1 \tensor S) \circ \Delta = unit \circ counit

            - `S` is an *anti*-homomorphism

            These properties are tested on :meth:`some_elements`.

            TESTS::

                sage: R = NonCommutativeSymmetricFunctions(QQ).ribbon()                 # needs sage.combinat sage.modules
                sage: R._test_antipode()                                                # needs sage.combinat sage.modules

            ::

                sage: s = SymmetricFunctions(QQ).schur()                                # needs sage.combinat sage.modules
                sage: s._test_antipode()                                                # needs lrcalc_python sage.combinat sage.modules
            """
            tester = self._tester(**options)

            S = self.antipode

            IS = lambda x: self.sum(c * self.monomial(t1) * S(self.monomial(t2))
                                for ((t1, t2), c) in x.coproduct())

            SI = lambda x: self.sum(c * S(self.monomial(t1)) * self.monomial(t2)
                                for ((t1, t2), c) in x.coproduct())

            for x in tester.some_elements():

                # antipode is an anti-homomorphism
                for y in tester.some_elements():
                    tester.assertEqual(S(x) * S(y), S(y * x))

                # mu * (S # I) * delta == counit * unit
                tester.assertEqual(SI(x), self.counit(x) * self.one())

                # mu * (I # S) * delta == counit * unit
                tester.assertEqual(IS(x), self.counit(x) * self.one())

    class ElementMethods:
        pass

    class TensorProducts(TensorProductsCategory):
        """
        The category of Hopf algebras with basis constructed by tensor product of Hopf algebras with basis
        """

        @cached_method
        def extra_super_categories(self):
            """
            EXAMPLES::

                sage: C = HopfAlgebrasWithBasis(QQ).TensorProducts()
                sage: C.extra_super_categories()
                [Category of Hopf algebras with basis over Rational Field]
                sage: sorted(C.super_categories(), key=str)
                [Category of Hopf algebras with basis over Rational Field,
                 Category of tensor products of Hopf algebras over Rational Field,
                 Category of tensor products of algebras with basis over Rational Field]
            """
            return [self.base_category()]

        class ParentMethods:
            # todo: antipode
            pass

        class ElementMethods:
            pass



# sage_setup: distribution = sagemath-categories
r"""
Unital algebras
"""
# ****************************************************************************
#  Copyright (C) 2011 Nicolas M. Thiery <nthiery at users.sf.net>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  https://www.gnu.org/licenses/
# *****************************************************************************
from sage.misc.abstract_method import abstract_method
from sage.misc.cachefunc import cached_method
from sage.misc.lazy_attribute import lazy_attribute
from sage.categories.category import Category
from sage.categories.category_with_axiom import CategoryWithAxiom_over_base_ring
from sage.categories.commutative_additive_groups import CommutativeAdditiveGroups
from sage.categories.magmas import Magmas
from sage.categories.morphism import SetMorphism
from sage.categories.homset import Hom
from sage.categories.rings import Rings
from sage.categories.magmatic_algebras import MagmaticAlgebras
from sage.categories.cartesian_product import CartesianProductsCategory


class UnitalAlgebras(CategoryWithAxiom_over_base_ring):
    """
    The category of non-associative algebras over a given base ring.

    A non-associative algebra over a ring `R` is a module over `R`
    which is also a unital magma.

    .. WARNING::

        Until :issue:`15043` is implemented, :class:`Algebras` is the
        category of associative unital algebras; thus, unlike the name
        suggests, :class:`UnitalAlgebras` is not a subcategory of
        :class:`Algebras` but of
        :class:`~.magmatic_algebras.MagmaticAlgebras`.

    EXAMPLES::

        sage: from sage.categories.unital_algebras import UnitalAlgebras
        sage: C = UnitalAlgebras(ZZ); C
        Category of unital algebras over Integer Ring

    TESTS::

        sage: from sage.categories.magmatic_algebras import MagmaticAlgebras
        sage: C is MagmaticAlgebras(ZZ).Unital()
        True
        sage: TestSuite(C).run()
    """
    _base_category_class_and_axiom = (MagmaticAlgebras, "Unital")

    class ParentMethods:
        def from_base_ring(self, r):
            """
            Return the canonical embedding of ``r`` into ``self``.

            INPUT:

            - ``r`` -- an element of ``self.base_ring()``

            EXAMPLES::

                sage: A = AlgebrasWithBasis(QQ).example(); A                            # needs sage.combinat sage.modules
                An example of an algebra with basis:
                 the free algebra on the generators ('a', 'b', 'c') over Rational Field
                sage: A.from_base_ring(1)                                               # needs sage.combinat sage.modules
                B[word: ]
            """
            return self.one()._lmul_(r)

        def __init_extra__(self):
            """
            Declare the canonical coercion from ``self.base_ring()``
            to ``self``, if there has been none before.

            EXAMPLES::

                sage: A = AlgebrasWithBasis(QQ).example(); A                            # needs sage.combinat sage.modules
                An example of an algebra with basis:
                 the free algebra on the generators ('a', 'b', 'c') over Rational Field
                sage: coercion_model = sage.structure.element.get_coercion_model()
                sage: coercion_model.discover_coercion(QQ, A)                           # needs sage.combinat sage.modules
                ((map internal to coercion system -- copy before use)
                 Generic morphism:
                   From: Rational Field
                   To:   An example of an algebra with basis:
                          the free algebra on the generators ('a', 'b', 'c') over Rational Field,
                 None)
                sage: A(1)          # indirect doctest                                  # needs sage.combinat sage.modules
                B[word: ]

            TESTS:

            Ensure that :issue:`28328` is fixed and that non-associative
            algebras are supported::

                sage: # needs sage.modules
                sage: class Foo(CombinatorialFreeModule):
                ....:     def one(self):
                ....:         return self.monomial(0)
                sage: from sage.categories.magmatic_algebras import MagmaticAlgebras
                sage: C = MagmaticAlgebras(QQ).WithBasis().Unital()
                sage: F = Foo(QQ, (1,), category=C)
                sage: F(0)
                0
                sage: F(3)
                3*B[0]
            """
            base_ring = self.base_ring()
            if base_ring is self:
                # There are rings that are their own base rings. No need to register that.
                return
            if self._is_coercion_cached(base_ring):
                # We will not use any generic stuff, since a (presumably) better conversion
                # has already been registered.
                return
            if base_ring is None:
                # It may happen that self.base_ring() is not initialised at this point.
                return

            mor = self._coerce_map_from_base_ring()
            if mor is not None:
                mor._make_weak_references()
                try:
                    self.register_coercion(mor)
                except AssertionError:
                    pass

        def _coerce_map_from_(self, other):
            """
            Return a coercion map from ``other`` to ``self``, or ``None``.

            TESTS:

            Check that :issue:`19225` is solved::

                sage: A = cartesian_product((QQ['z'],)); A
                The Cartesian product of (Univariate Polynomial Ring in z over Rational Field,)
                sage: A.coerce_map_from(ZZ)
                Composite map:
                  From: Integer Ring
                  To:   The Cartesian product of (Univariate Polynomial Ring in z over Rational Field,)
                  Defn:   Natural morphism:
                          From: Integer Ring
                          To:   Rational Field
                        then
                          Generic morphism:
                          From: Rational Field
                          To:   The Cartesian product of (Univariate Polynomial Ring in z over Rational Field,)
                sage: A(1)
                (1,)
            """
            if other is self.base_ring():
                return self._coerce_map_from_base_ring()
            else:
                return self._coerce_map_via([self.base_ring()], other)

        def _coerce_map_from_base_ring(self):
            """
            Return a suitable coercion map from the base ring of ``self``.

            TESTS::

                sage: A = cartesian_product((QQ['z'],)); A
                The Cartesian product of (Univariate Polynomial Ring in z over Rational Field,)
                sage: A.base_ring()
                Rational Field
                sage: A._coerce_map_from_base_ring()
                Generic morphism:
                From: Rational Field
                To:   The Cartesian product of (Univariate Polynomial Ring in z over Rational Field,)

            Check that :issue:`29312` is fixed::

                sage: F.<x,y,z> = FreeAlgebra(QQ, implementation='letterplace')         # needs sage.combinat sage.modules
                sage: F._coerce_map_from_base_ring()                                    # needs sage.combinat sage.modules
                Generic morphism:
                  From: Rational Field
                  To:   Free Associative Unital Algebra on 3 generators (x, y, z) over Rational Field
            """
            base_ring = self.base_ring()

            # Pick a homset for the morphism to live in...
            if self in Rings():
                # The algebra is associative, and thus a ring. The
                # base ring is also a ring. Everything is OK.
                H = Hom(base_ring, self, Rings())
            else:
                # If the algebra isn't associative, we would like to
                # use the category of unital magmatic algebras (which
                # are not necessarily associative) instead. But,
                # unfortunately, certain important rings like QQ
                # aren't in that category. As a result, we have to use
                # something weaker.
                cat = Magmas().Unital()
                cat = Category.join([cat, CommutativeAdditiveGroups()])
                cat = cat.Distributive()
                H = Hom(base_ring, self, cat)

            # We need to construct a coercion from the base ring to self.
            #
            # There is a generic method from_base_ring(), that just does
            # multiplication with the multiplicative unit. However, the
            # unit is constructed repeatedly, which is slow.
            # So, if the unit is available *now*, then we can create a
            # faster coercion map.
            #
            # This only applies for the generic from_base_ring() method.
            # If there is a specialised from_base_ring(), then it should
            # be used unconditionally.
            generic_from_base_ring = self.category().parent_class.from_base_ring
            from_base_ring = self.from_base_ring   # bound method
            if from_base_ring.__func__ != generic_from_base_ring:
                # Custom from_base_ring()
                use_from_base_ring = True
            elif isinstance(generic_from_base_ring, lazy_attribute):
                # If the category implements from_base_ring() as lazy
                # attribute, then we always use it.
                # This is for backwards compatibility, see Issue #25181
                use_from_base_ring = True
            else:
                try:
                    one = self.one()
                    use_from_base_ring = False
                except (NotImplementedError, AttributeError, TypeError):
                    # The unit is not available, yet. But there are cases
                    # in which it will be available later. So, we use
                    # the generic from_base_ring() after all.
                    use_from_base_ring = True

            mor = None
            if use_from_base_ring:
                mor = SetMorphism(function=from_base_ring, parent=H)
            else:
                # We have the multiplicative unit, so implement the
                # coercion from the base ring as multiplying with that.
                #
                # But first we check that it actually works. If not,
                # then the generic implementation of from_base_ring()
                # would fail as well so we don't use it.
                try:
                    if one._lmul_(base_ring.an_element()) is not None:
                        # There are cases in which lmul returns None,
                        # which means that it's not implemented.
                        # One example: Hecke algebras.
                        mor = SetMorphism(function=one._lmul_, parent=H)
                except (NotImplementedError, AttributeError, TypeError):
                    pass
            return mor

    class WithBasis(CategoryWithAxiom_over_base_ring):

        class ParentMethods:

            @abstract_method(optional=True)
            def one_basis(self):
                """
                When the one of an algebra with basis is an element of
                this basis, this optional method can return the index of
                this element. This is used to provide a default
                implementation of :meth:`.one`, and an optimized default
                implementation of :meth:`.from_base_ring`.

                EXAMPLES::

                    sage: # needs sage.combinat sage.modules
                    sage: A = AlgebrasWithBasis(QQ).example()
                    sage: A.one_basis()
                    word:
                    sage: A.one()
                    B[word: ]
                    sage: A.from_base_ring(4)
                    4*B[word: ]
                """

            @cached_method
            def one_from_one_basis(self):
                """
                Return the one of the algebra, as per
                :meth:`Monoids.ParentMethods.one()
                <sage.categories.monoids.Monoids.ParentMethods.one>`

                By default, this is implemented from
                :meth:`.one_basis`, if available.

                EXAMPLES::

                    sage: # needs sage.combinat sage.modules
                    sage: A = AlgebrasWithBasis(QQ).example()
                    sage: A.one_basis()
                    word:
                    sage: A.one_from_one_basis()
                    B[word: ]
                    sage: A.one()
                    B[word: ]

                TESTS:

                Try to check that :issue:`5843` Heisenbug is fixed::

                    sage: # needs sage.combinat sage.modules
                    sage: A = AlgebrasWithBasis(QQ).example()
                    sage: B = AlgebrasWithBasis(QQ).example(('a', 'c'))
                    sage: A == B
                    False
                    sage: Aone = A.one_from_one_basis
                    sage: Bone = B.one_from_one_basis
                    sage: Aone is Bone
                    False

                Even if called in the wrong order, they should returns their
                respective one::

                    sage: Bone().parent() is B                                          # needs sage.combinat sage.modules
                    True
                    sage: Aone().parent() is A                                          # needs sage.combinat sage.modules
                    True
                """
                return self.monomial(self.one_basis()) #.

            @lazy_attribute
            def one(self):
                r"""
                Return the multiplicative unit element.

                EXAMPLES::

                    sage: A = AlgebrasWithBasis(QQ).example()                           # needs sage.combinat sage.modules
                    sage: A.one_basis()                                                 # needs sage.combinat sage.modules
                    word:
                    sage: A.one()                                                       # needs sage.combinat sage.modules
                    B[word: ]
                """
                if self.one_basis is NotImplemented:
                    return NotImplemented
                return self.one_from_one_basis

            @lazy_attribute
            def from_base_ring(self):
                """
                TESTS::

                    sage: A = AlgebrasWithBasis(QQ).example()                           # needs sage.combinat sage.modules
                    sage: A.from_base_ring(3)                                           # needs sage.combinat sage.modules
                    3*B[word: ]
                """
                if self.one_basis is NotImplemented:
                    return NotImplemented
                return self.from_base_ring_from_one_basis

            def from_base_ring_from_one_basis(self, r):
                """
                Implement the canonical embedding from the ground ring.

                INPUT:

                - ``r`` -- an element of the coefficient ring

                EXAMPLES::

                    sage: # needs sage.combinat sage.modules
                    sage: A = AlgebrasWithBasis(QQ).example()
                    sage: A.from_base_ring_from_one_basis(3)
                    3*B[word: ]
                    sage: A.from_base_ring(3)
                    3*B[word: ]
                    sage: A(3)
                    3*B[word: ]
                """
                return self.term(self.one_basis(), r)

    class CartesianProducts(CartesianProductsCategory):
        r"""
        The category of unital algebras constructed as Cartesian products
        of unital algebras.

        This construction gives the direct product of algebras. See
        discussion on:

         - http://groups.google.fr/group/sage-devel/browse_thread/thread/35a72b1d0a2fc77a/348f42ae77a66d16#348f42ae77a66d16
         - :wikipedia:`Direct_product`
        """
        def extra_super_categories(self):
            """
            A Cartesian product of algebras is endowed with a natural
            unital algebra structure.

            EXAMPLES::

                sage: from sage.categories.unital_algebras import UnitalAlgebras
                sage: C = UnitalAlgebras(QQ).CartesianProducts()
                sage: C.extra_super_categories()
                [Category of unital algebras over Rational Field]
                sage: sorted(C.super_categories(), key=str)
                [Category of Cartesian products of distributive magmas and additive magmas,
                 Category of Cartesian products of unital magmas,
                 Category of Cartesian products of vector spaces over Rational Field,
                 Category of unital algebras over Rational Field]
            """
            return [self.base_category()]

        class ParentMethods:
            @cached_method
            def one(self):
                r"""
                Return the multiplicative unit element.

                EXAMPLES::

                    sage: # needs sage.graphs sage.modules
                    sage: S2 = simplicial_complexes.Sphere(2)
                    sage: H = S2.cohomology_ring(QQ)
                    sage: C = cartesian_product([H, H])
                    sage: one = C.one()
                    sage: one
                    B[(0, (0, 0))] + B[(1, (0, 0))]
                    sage: one == one * one
                    True
                    sage: all(b == b * one for b in C.basis())
                    True
                """
                data = enumerate(self.cartesian_factors())
                return self._cartesian_product_of_elements([C.one() for i, C in data])



# sage_setup: distribution = sagemath-categories
r"""
Rngs
"""
#*****************************************************************************
#  Copyright (C) 2008 Teresa Gomez-Diaz (CNRS) <Teresa.Gomez-Diaz@univ-mlv.fr>
#                2012 Nicolas M. Thiery <nthiery at users.sf.net>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
#******************************************************************************

from sage.categories.category_with_axiom import CategoryWithAxiom
from sage.misc.lazy_import import LazyImport
from sage.categories.magmas_and_additive_magmas import MagmasAndAdditiveMagmas


class Rngs(CategoryWithAxiom):
    """
    The category of rngs.

    An *rng* `(S, +, *)` is similar to a ring but not necessarily
    unital. In other words, it is a combination of a commutative
    additive group `(S, +)` and a multiplicative semigroup `(S, *)`,
    where `*` distributes over `+`.

    EXAMPLES::

        sage: C = Rngs(); C
        Category of rngs
        sage: sorted(C.super_categories(), key=str)
        [Category of associative additive commutative additive associative additive unital distributive magmas and additive magmas,
         Category of commutative additive groups]

        sage: sorted(C.axioms())
        ['AdditiveAssociative', 'AdditiveCommutative', 'AdditiveInverse',
         'AdditiveUnital', 'Associative', 'Distributive']

        sage: C is (CommutativeAdditiveGroups() & Semigroups()).Distributive()
        True
        sage: C.Unital()
        Category of rings

    TESTS::

        sage: TestSuite(C).run()
    """

    _base_category_class_and_axiom = (MagmasAndAdditiveMagmas.Distributive.AdditiveAssociative.AdditiveCommutative.AdditiveUnital.Associative, "AdditiveInverse")

    Unital = LazyImport('sage.categories.rings', 'Rings', at_startup=True)
